"""
#=========================================================================
# OpenLoopCLPass.py
#=========================================================================
# Generate a simple schedule (no Mamba techniques here) based on the
# DAG generated by some previous pass.
#
# Author : Shunning Jiang
# Date   : Apr 20, 2019
"""
import os
import random
from collections import deque

import py

from pymtl3.datatypes import Bits1
from pymtl3.dsl import CalleeIfcCL, CalleePort
from pymtl3.dsl.errors import UpblkCyclicError

from ..BasePass import BasePass, PassMetadata
from ..errors import PassOrderError
from ..sim.SimpleSchedulePass import SimpleSchedulePass, dump_dag
from ..sim.SimpleTickPass import SimpleTickPass
from ..tracing.CLLineTracePass import CLLineTracePass
from ..tracing.CollectSignalPass import CollectSignalPass
from ..tracing.PrintWavePass import PrintWavePass
from ..tracing.VcdGenerationPass import VcdGenerationPass

random.seed(0xdeadbeef)

class OpenLoopCLPass( BasePass ):
  def __init__( self, print_line_trace=True ):
    self.line_trace_on = print_line_trace

  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    top._sched = PassMetadata()

    self.schedule_with_top_level_callee( top )

  def schedule_with_top_level_callee( self, top ):

    # Construct the graph with top level callee port
    V = top._dag.final_upblks - top.get_all_update_ff()
    E = set()

    # We collect all top level callee ports/nonblocking callee interfaces
    top_level_callee_ports = top.get_all_object_filter(
      lambda x: isinstance(x, CalleePort) and x.get_host_component() is top )

    top_level_nb_ifcs = top.get_all_object_filter(
      lambda x: isinstance(x, CalleeIfcCL) and x.get_host_component() is top )

    method_callee_mapping = {}
    method_guard_mapping  = {}
    guard_method_mapping  = {}

    def get_raw_method( x ):
      assert isinstance( x, CalleePort )
      return x.method

    # First deal with normal calleeports. We map the actual method to the
    # callee port, and add the port to the vertex set
    for x in top_level_callee_ports:
      if not x.in_non_blocking_interface(): # Normal callee port
        V.add(x)
        m = get_raw_method( x )
        assert m not in method_callee_mapping
        method_callee_mapping[m] = x

    # Then deal with non-blocking callee interfaces. Map the method of the
    # interface to the actual method and set up method-rdy mapping
    for x in top_level_nb_ifcs:
      V.add( x.method )
      V.add( x.rdy )
      E.add( (x.rdy, x.method) )

      method_guard_mapping[x.method] = x.rdy
      guard_method_mapping[x.rdy] = x.method
      m = get_raw_method( x.method )
      r = get_raw_method( x.rdy )

      assert m not in method_callee_mapping
      method_callee_mapping[m] = x.method
      method_callee_mapping[r] = x.rdy

    G   = { v: [] for v in V }
    G_T = { v: [] for v in V }

    for (u, v) in top._dag.all_constraints:
      if u in V and v in V:
        E.add( (u, v) )
        G  [u].append( v )
        G_T[v].append( u )

    # In addition to existing constraints, we process the constraints that
    # involve top level callee ports. NOTE THAT we assume the user never
    # set the constraint on the actual method object inside the CalleePort
    # In GenDAGPass we already collect those constraints between update
    # blocks and ACTUAL METHODs. We use the ACTUAL METHOD to callee
    # mapping we set up above to avoid missing constraints.

    for (xx, yy) in top._dag.top_level_callee_constraints:

      if xx in method_callee_mapping:
        xx = method_callee_mapping[ xx ]

      if yy in method_callee_mapping:
        yy = method_callee_mapping[ yy ]

      # if there is a constraint A<B, it means M(A.method) < B.rdy or means U(A) < B.rdy
      if yy in method_guard_mapping:
        yy = method_guard_mapping[ yy ]

      if xx in V and yy in V:
        E.add( (xx, yy) )
        G  [xx].append( yy )
        G_T[yy].append( xx )

    if 'MAMBA_DAG' in os.environ:
      dump_dag( top, V, E )

    #---------------------------------------------------------------------
    # Run Kosaraju's algorithm to shrink all strongly connected components
    # (SCCs) into super nodes
    #---------------------------------------------------------------------
    # See passes/sim/DynamicSchedulePass.py
    # TODO refactor this

    PO = []

    vertices = list(G.keys())
    random.shuffle(vertices)
    visited = set()

    for u in vertices:
      stack = [ (u, False) ]
      while stack:
        u, second_visit = stack.pop()

        if second_visit:
          PO.append( u )
        elif u not in visited:
          visited.add( u )
          stack.append( (u, True) )
          for v in reversed(G[u]):
            stack.append( (v, False) )

    RPO = PO[::-1]

    # Second bfs on G_T to generate SCCs

    SCCs  = []
    v_SCC = {}
    visited = set()

    for u in RPO:
      if u not in visited:
        visited.add( u )
        scc = set()
        SCCs.append( scc )
        Q = deque( [u] )
        scc.add( u )
        while Q:
          u = Q.popleft()
          v_SCC[u] = len(SCCs) - 1
          for v in G_T[u]:
            if v not in visited:
              visited.add( v )
              Q.append( v )
              scc.add( v )

    # Construct a new graph of SCCs

    G_new = { i: set() for i in range(len(SCCs)) }
    InD   = { i: 0     for i in range(len(SCCs)) }

    for (u, v) in E: # u -> v
      scc_u, scc_v = v_SCC[u], v_SCC[v]
      if scc_u != scc_v and scc_v not in G_new[ scc_u ]:
        InD[ scc_v ] += 1
        G_new[ scc_u ].add( scc_v )

    # Perform topological sort on SCCs

    scc_pred = {}
    scc_schedule = []

    Q = list( [ i for i in range(len(SCCs)) if not InD[i] ] )
    for x in Q:
      scc_pred[ x ] = None

    list_version_of_SCCs = [ list(x) for x in SCCs ]
    while Q:
      # random.shuffle(Q)

      # Prioritize update blocks instead of method
      # TODO make it O(logn) by balanced BST if needed ...
      found = False
      for i in range(len(Q)):
        m = Q[i]
        if m not in method_guard_mapping or m not in guard_method_mapping:
          u = Q.pop(i)
          found = True
          break

      if found:
        scc_schedule.append( u )

      for v in G_new[u]:
        InD[v] -= 1
        if not InD[v]:
          Q.append( v )
          scc_pred[ v ] = u

    assert len(scc_schedule) == len(SCCs), "{} != {}".format(len(scc_schedule), len(SCCs))

    #---------------------------------------------------------------------
    # Now we generate super blocks for each SCC and produce final schedule
    #---------------------------------------------------------------------

    constraint_objs = top._dag.constraint_objs

    update_schedule = []

    scc_id = 0
    for i in scc_schedule:
      scc = SCCs[i]
      if len(scc) == 1:
        u = list(scc)[0]
        update_schedule.append( u )
      else:

        # For each non-trivial SCC, we need to figure out a intra-SCC
        # linear schedule that minimizes the time to re-execute this SCC
        # due to value changes. A bad schedule may inefficiently execute
        # the SCC for many times, each of which changes a few signals.
        # The current algorithm iteratively finds the "entry block" of
        # the SCC and expand its adjancent blocks. The implementation is
        # to first find the actual entry point, and then BFS to expand the
        # footprint until all nodes are visited.

        tmp_schedule = []
        Q = deque()

        if scc_pred[i] is None:
          # We start bfs from the block that has the least number of input
          # edges in the SCC
          InD = { v: 0 for v in scc }
          for (u, v) in E: # u -> v
            if u in scc and v in scc:
              InD[ v ] += 1
          Q.append( max(InD, key=InD.get) )

        else:
          # We start bfs with the blocks that are successors of the
          # predecessor scc in the previous SCC-level topological sort.
          pred = set( SCCs[ scc_pred[i] ] )
          # Sort by names for a fixed outcome
          for x in sorted( scc, key = lambda x: x.__name__ ):
            for v in G_T[x]: # find reversed edges point back to pred SCC
              if v in pred:
                Q.append( x )

        # Perform bfs to find a heuristic schedule
        visited = set(Q)
        while Q:
          u = Q.popleft()
          tmp_schedule.append( u )
          for v in G[u]:
            if v in scc and v not in visited:
              Q.append( v )
              visited.add( v )

        scc_id += 1
        variables = set()
        for (u, v) in E:
          # Collect all variables that triggers other blocks in the SCC
          if u in scc and v in scc:
            variables.update( constraint_objs[ (u, v) ] )

        # generate a loop for scc
        # Shunning: we just simply loop over the whole SCC block
        # TODO performance optimizations using Mamba techniques within a SCC block

        def gen_wrapped_SCCblk( s, scc, src ):
          from pymtl3.dsl.errors import UpblkCyclicError

          namespace = {}
          namespace.update( locals() )
          # print src
          exec(py.code.Source( src ).compile(), namespace)

          return namespace['generated_block']

        # FIXME when there is nothing in {2} ..
        template = """
          from copy import deepcopy
          def wrapped_SCC_{0}():
            num_iters = 0
            while True:
              num_iters += 1
              {1}
              for blk in scc: # TODO Mamba
                blk()
              if {2}:
                break
              if num_iters > 100:
                raise UpblkCyclicError("Combinational loop detected at runtime in {{{3}}}!")
            # print "SCC block{0} is executed", num_iters, "times"
          generated_block = wrapped_SCC_{0}
        """

        copy_srcs  = []
        check_srcs = []
        # print_srcs = []

        for j, var in enumerate(variables):
          if issubclass( var._dsl.Type, Bits ):
            copy_srcs.append( f"t{j} = {var}.value" )
          elif is_bitstruct_class( var._dsl.Type ):
            copy_srcs.append( f"t{j} = {var}.clone()" )
          else:
            copy_srcs.append( f"t{j} = deepcopy({var})" )

          check_srcs.append( f"{var} == t{j}" )
          # print_srcs.append( f"print '{var}', {var}, _____tmp_{j}" )

        scc_block_src = template.format( scc_id,
                                         "; ".join( copy_srcs ),
                                         " and ".join( check_srcs ),
                                         ", ".join( [ x.__name__ for x in scc] ) )
                                         # "; ".join( print_srcs ) )

        update_schedule.append( gen_wrapped_SCCblk( top, tmp_schedule, scc_block_src ) )

    # Shunning: we call line trace related pass here.
    CLLineTracePass()( top )
    CollectSignalPass()( top )
    VcdGenerationPass()( top )
    PrintWavePass()( top )

    # Shunning: we reuse ff and posedge schedules from SimpleSchedulePass
    simple = SimpleSchedulePass()
    simple.schedule_ff( top )
    simple.schedule_posedge_flip( top )

    # Currently the tick order is:
    # [ clear_cl_trace, update, ff, tracing, posedge ]
    # in order to avoid ticking for the first cycle.

    schedule = []

    # clear cl method flag
    schedule.append( top._tracing.clear_cl_trace )

    # execute all update blocks
    schedule.extend( update_schedule )

    # print trace after all update blocks
    def print_line_trace():
      print(top.num_cycles_executed, top.__class__.__name__.ljust(15), ':', top.line_trace())

    if self.line_trace_on:
      schedule.append( print_line_trace )

    # call ff blocks first
    schedule.extend( top._sched.schedule_ff )

    # append tracing related work

    if hasattr( top, "_tracing" ):
      if hasattr( top._tracing, "vcd_func" ):
        schedule.append( top._tracing.vcd_func )
      if hasattr( top._tracing, "collect_text_sigs" ):
        schedule.append( top._tracing.collect_text_sigs )

    # posedge flip
    schedule.extend( top._sched.schedule_posedge_flip )

    top._sched.new_schedule_index  = 0
    top._sched.orig_schedule_index = 0


    # Here we are trying to avoid scanning the original schedule that
    # contains methods because we will need isinstance in that case.
    # As a result we created a preprocessed list for execution and use
    # the dictionary to look up the new index of functions.

    schedule_no_method = [ x for x in schedule if not isinstance(x, CalleePort) ]
    mapping = { x : i for i, x in enumerate( schedule_no_method ) }

    def wrap_method( top, method,
                     my_idx_new, schedule_no_method,
                     my_idx_orig ):

      def actual_method( *args, **kwargs ):
        i = top._sched.new_schedule_index
        j = top._sched.orig_schedule_index

        if j > my_idx_orig:
          # This means we need to advance the current cycle to the end
          # and then normally execute until we get to the same point.
          # We use original schedule index to handle the case where
          # there are two consecutive methods.
          while i < len(schedule_no_method):
            schedule_no_method[i]()
            i += 1
          i = j = 0
          top.num_cycles_executed += 1

        # We advance from the current point i to the method's position in
        # the schedule without method just to execute those blocks
        while i < my_idx_new:
          schedule_no_method[i]()
          i += 1
        j = my_idx_orig + 1

        # Execute the method
        ret = method( *args, **kwargs )

        top._sched.new_schedule_index = i
        top._sched.orig_schedule_index = j
        return ret

      return actual_method

    for i, x in enumerate( schedule ):
      if isinstance( x, CalleePort ):
        x.original_method = x.method

        # This is to find the next non-method block's position in the
        # original schedule
        next_func   = i + 1
        while next_func < len(schedule):
          if not isinstance( schedule[next_func], CalleePort ):
            break
          next_func += 1

        # Get the index of the block in the schedule without method
        # This always exists because we append a line trace at the end
        map_next_func = mapping[ schedule[next_func] ]

        x.method = wrap_method( top, x.method,
                                map_next_func,
                                schedule_no_method,
                                i )
    top.num_cycles_executed = 0

    # This is for reset to work correctly
    top.tick = SimpleTickPass.gen_tick_function( schedule_no_method )



from .datatypes import *
from .datatypes import _bitwidths
from .dsl.Component import Component
from .dsl.ComponentLevel3 import connect
from .dsl.ComponentLevel5 import method_port
from .dsl.ComponentLevel6 import non_blocking
from .dsl.ComponentLevel7 import blocking
from .dsl.Connectable import (
    CalleeIfcCL,
    CalleeIfcFL,
    CalleeIfcRTL,
    CalleePort,
    CallerIfcCL,
    CallerIfcFL,
    CallerIfcRTL,
    CallerPort,
    InPort,
    Interface,
    OutPort,
    Wire,
)
from .dsl.ConstraintTypes import RD, WR, M, U
from .dsl.Placeholder import Placeholder
from .passes import ImportConfigs, TracingConfigs
from .passes.backends.sverilog import ImportPass, TranslationImportPass
from .passes.PassGroups import SimulationPass

__version__ = "0.5.5"

__all__ = [
  'U','M','RD','WR',
  'Wire', 'InPort', 'OutPort', 'Interface', 'CallerPort', 'CalleePort',
  'connect', 'method_port',
  'CalleeIfcRTL', 'CallerIfcRTL',
  'non_blocking', 'CalleeIfcCL', 'CallerIfcCL',
  'blocking', 'CalleeIfcFL', 'CallerIfcFL',

  'SimulationPass', 'TracingConfigs', 'TranslationImportPass', 'ImportPass',
  'Component', 'Placeholder',

  'sext', 'zext', 'clog2', 'concat', 'reduce_and', 'reduce_or', 'reduce_xor',
  'mk_bits', 'Bits',
  'mk_bitstruct', 'bitstruct',
  'to_bits', 'get_nbits',
] + [ "Bits{}".format(x) for x in _bitwidths ] \
  + [ "b{}".format(x) for x in _bitwidths ]



"""Expose structural RTLIR generation pass.

PyMTL user should only interact with the passes exposed here.
"""
from .StructuralRTLIRGenL4Pass import StructuralRTLIRGenL4Pass as StructuralRTLIRGenPass



from pymtl3 import *
from pymtl3.stdlib.ifcs import InValRdyIfc, OutValRdyIfc
from pymtl3.stdlib.rtl import Mux, Reg, RegEn, RegisterFile


class PipeQueue1RTL( Component ):

  def construct( s, Type ):
    s.enq = InValRdyIfc ( Type )
    s.deq = OutValRdyIfc( Type )

    s.buffer  = RegEn( Type )( out = s.deq.msg, in_ = s.enq.msg )

    s.next_full = Wire( Bits1 )
    s.full      = Wire( Bits1 )
    connect( s.full, s.deq.val )

    @s.update_ff
    def up_full():
      s.full <<= s.next_full

    @s.update
    def up_pipeq_set_enq_rdy():
      s.enq.rdy = ~s.full | s.deq.rdy

    @s.update
    def up_pipeq_full():
      s.buffer.en = s.enq.val & s.enq.rdy
      s.next_full = s.enq.val | (s.full & ~s.deq.rdy)

  def line_trace( s ):
    return s.enq.line_trace() + " > " + \
            ("*" if s.full else " ") + " > " + \
            s.deq.line_trace()

class BypassQueue1RTL( Component ):

  def construct( s, Type ):
    s.enq = InValRdyIfc ( Type )
    s.deq = OutValRdyIfc( Type )

    s.buffer = RegEn( Type )( in_ = s.enq.msg )

    s.next_full = Wire( Bits1 )
    s.full      = Wire( Bits1 )

    s.byp_mux = Mux( Type, 2 )(
      out = s.deq.msg,
      in_ = { 0: s.enq.msg,
              1: s.buffer.out, },
      sel = s.full, # full -- buffer.out, empty -- bypass
    )

    @s.update_ff
    def up_full():
      s.full <<= s.next_full

    @s.update
    def up_bypq_set_enq_rdy():
      s.enq.rdy = ~s.full

    @s.update
    def up_bypq_internal():
      s.buffer.en = (~s.deq.rdy) & (s.enq.val & s.enq.rdy)
      s.next_full = (~s.deq.rdy) & s.deq.val

    # this enables the sender to make enq.val depend on enq.rdy
    @s.update
    def up_bypq_set_deq_val():
      s.deq.val = s.full | s.enq.val

  def line_trace( s ):
    return s.enq.line_trace() + " > " + \
            ("*" if s.full else " ") + " > " + \
            s.deq.line_trace()

class NormalQueue1RTL( Component ):

  def construct( s, Type ):
    s.enq = InValRdyIfc( Type )
    s.deq = OutValRdyIfc( Type )

    s.buffer  = RegEn( Type )( out = s.deq.msg, in_ = s.enq.msg )

    s.next_full = Wire( Bits1 )
    s.full      = Wire( Bits1 )
    connect( s.full, s.deq.val )

    @s.update_ff
    def up_full():
      s.full <<= s.next_full

    @s.update
    def up_normq_set_enq_rdy():
      s.enq.rdy = ~s.full

    @s.update
    def up_normq_internal():
      s.buffer.en = s.enq.val & s.enq.rdy
      s.next_full = (s.full & ~s.deq.rdy) | s.buffer.en

  def line_trace( s ):
    return s.enq.line_trace() + " > " + \
            ("*" if s.full else " ") + " > " + \
            s.deq.line_trace()

#-----------------------------------------------------------------------
# NormalQueueRTL
#-----------------------------------------------------------------------
class NormalQueueRTL( Component ):

  def construct( s, num_entries, Type ):

    s.enq              = InValRdyIfc( Type )
    s.deq              = OutValRdyIfc( Type )
    s.num_free_entries = OutPort( mk_bits( clog2(num_entries) ) )

    # Ctrl and Dpath unit instantiation

    s.ctrl  = NormalQueueRTLCtrl ( num_entries       )
    s.dpath = NormalQueueRTLDpath( num_entries, Type )

    # Ctrl unit connections

    connect( s.ctrl.enq_val,          s.enq.val          )
    connect( s.ctrl.enq_rdy,          s.enq.rdy          )
    connect( s.ctrl.deq_val,          s.deq.val          )
    connect( s.ctrl.deq_rdy,          s.deq.rdy          )
    connect( s.ctrl.num_free_entries, s.num_free_entries )

    # Dpath unit connections

    connect( s.dpath.enq_bits, s.enq.msg    )
    connect( s.dpath.deq_bits, s.deq.msg    )

    # Control Signal connections (ctrl -> dpath)

    connect( s.dpath.wen,      s.ctrl.wen   )
    connect( s.dpath.waddr,    s.ctrl.waddr )
    connect( s.dpath.raddr,    s.ctrl.raddr )

  def line_trace( s ):
    return "{} () {}".format( s.enq, s.deq )

#-----------------------------------------------------------------------
# NormalQueueRTLDpath
#-----------------------------------------------------------------------
class NormalQueueRTLDpath( Component ):

  def construct( s, num_entries, Type ):

    AddrType    = mk_bits( clog2( num_entries ) )

    s.enq_bits  = InPort  ( Type )
    s.deq_bits  = OutPort ( Type )

    # Control signal (ctrl -> dpath)
    s.wen       = InPort  ( Bits1 )
    s.waddr     = InPort  ( AddrType )
    s.raddr     = InPort  ( AddrType )

    # Queue storage

    s.queue = RegisterFile( Type, num_entries )

    # Connect queue storage

    connect( s.queue.raddr[0], s.raddr    )
    connect( s.queue.rdata[0], s.deq_bits )
    connect( s.queue.wen[0],   s.wen      )
    connect( s.queue.waddr[0], s.waddr    )
    connect( s.queue.wdata[0], s.enq_bits )

#-----------------------------------------------------------------------
# NormalQueueRTLCtrl
#-----------------------------------------------------------------------
class NormalQueueRTLCtrl( Component ):

  def construct( s, num_entries ):

    AddrType      = mk_bits( clog2( num_entries ) )

    s.num_entries = AddrType( num_entries )

    # Interface Ports

    s.enq_val          = InPort  ( Bits1 )
    s.enq_rdy          = OutPort ( Bits1 )
    s.deq_val          = OutPort ( Bits1 )
    s.deq_rdy          = InPort  ( Bits1 )
    s.num_free_entries = OutPort ( AddrType )

    # Control signal (ctrl -> dpath)
    s.wen              = OutPort ( Bits1 )
    s.waddr            = OutPort ( AddrType )
    s.raddr            = OutPort ( AddrType )

    # Wires

    s.full             = Wire ( Bits1 )
    s.empty            = Wire ( Bits1 )
    s.do_enq           = Wire ( Bits1 )
    s.do_deq           = Wire ( Bits1 )
    s.enq_ptr          = Wire ( AddrType )
    s.deq_ptr          = Wire ( AddrType )
    s.enq_ptr_next     = Wire ( AddrType )
    s.deq_ptr_next     = Wire ( AddrType )
    # TODO: can't infer these temporaries due to if statement, fix
    s.enq_ptr_inc      = Wire ( AddrType )
    s.deq_ptr_inc      = Wire ( AddrType )
    s.full_next_cycle  = Wire ( Bits1 )

    s.last_idx         = AddrType( num_entries - 1 )

    @s.update
    def comb():

      # only enqueue/dequeue if valid and ready

      s.do_enq = s.enq_rdy and s.enq_val
      s.do_deq = s.deq_rdy and s.deq_val

      # write enable

      s.wen     = s.do_enq

      # enq ptr incrementer

      if s.enq_ptr == s.last_idx: s.enq_ptr_inc = AddrType(0)
      else:                       s.enq_ptr_inc = s.enq_ptr + AddrType(1)

      # deq ptr incrementer

      if s.deq_ptr == s.last_idx: s.deq_ptr_inc = AddrType(0)
      else:                       s.deq_ptr_inc = s.deq_ptr + AddrType(1)

      # set the next ptr value

      if s.do_enq: s.enq_ptr_next = s.enq_ptr_inc
      else:        s.enq_ptr_next = s.enq_ptr

      if s.do_deq: s.deq_ptr_next = s.deq_ptr_inc
      else:        s.deq_ptr_next = s.deq_ptr

      # number of free entries calculation

      if   s.reset:
        s.num_free_entries = s.num_entries
      elif s.full:
        s.num_free_entries = AddrType( 0 )
      elif s.empty:
        s.num_free_entries = s.num_entries
      elif s.enq_ptr > s.deq_ptr:
        s.num_free_entries = s.num_entries - ( s.enq_ptr - s.deq_ptr )
      elif s.deq_ptr > s.enq_ptr:
        s.num_free_entries = s.deq_ptr - s.enq_ptr

      s.full_next_cycle = (s.do_enq and not s.do_deq and
                                (s.enq_ptr_next == s.deq_ptr))

    @s.update
    def up_ctrl_signals():

      # set output signals

      s.empty   = not s.full and (s.enq_ptr == s.deq_ptr)

      s.enq_rdy = not s.full
      s.deq_val = not s.empty

      # set control signals

      s.waddr   = s.enq_ptr
      s.raddr   = s.deq_ptr

    @s.update_ff
    def seq():

      if s.reset:
        s.deq_ptr <<= AddrType( 0 )
        s.enq_ptr <<= AddrType( 0 )
      else:
        s.deq_ptr <<= s.deq_ptr_next
        s.enq_ptr <<= s.enq_ptr_next

      if   s.reset:               s.full <<= Bits1(0)
      elif s.full_next_cycle:     s.full <<= Bits1(1)
      elif (s.do_deq and s.full): s.full <<= Bits1(0)
      else:                       s.full <<= s.full



"""
=========================================================================
test_cases.py
=========================================================================
Centralized test case repository for the YosysVerilog backend.

Author : Peitian Pan
Date   : Dec 20, 2019
"""

from pymtl3.passes.backends.sverilog.testcases import (
    Bits32Foo,
    Bits32x5Foo,
    CaseArrayBits32IfcInUpblkComp,
    CaseBits32ArrayConnectSubCompAttrComp,
    CaseBits32ArraySubCompAttrUpblkComp,
    CaseBits32BitSelUpblkComp,
    CaseBits32ConnectSubCompAttrComp,
    CaseBits32FooInBits32OutComp,
    CaseBits32SubCompAttrUpblkComp,
    CaseBits32x2ConcatComp,
    CaseBits32x2ConcatConstComp,
    CaseBits32x2ConcatFreeVarComp,
    CaseBits32x2ConcatMixedComp,
    CaseBits32x2ConcatUnpackedSignalComp,
    CaseBits64PartSelUpblkComp,
    CaseBits64SextInComp,
    CaseBits64ZextInComp,
    CaseConnectArrayBits32FooIfcComp,
    CaseConnectArrayNestedIfcComp,
    CaseConnectArrayStructAttrToOutComp,
    CaseConnectArraySubCompArrayStructIfcComp,
    CaseConnectBitsConstToOutComp,
    CaseConnectBitSelToOutComp,
    CaseConnectConstStructAttrToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectInToWireComp,
    CaseConnectLiteralStructComp,
    CaseConnectNestedStructPackedArrayComp,
    CaseConnectPassThroughLongNameComp,
    CaseConnectSliceToOutComp,
    CaseConnectValRdyIfcComp,
    CaseConnectValRdyIfcUpblkComp,
    CaseConstStructInstComp,
    CaseElifBranchComp,
    CaseFixedSizeSliceComp,
    CaseForRangeLowerUpperStepPassThroughComp,
    CaseIfBasicComp,
    CaseIfBoolOpInForStmtComp,
    CaseIfDanglingElseInnerComp,
    CaseIfDanglingElseOutterComp,
    CaseIfExpInForStmtComp,
    CaseIfExpUnaryOpInForStmtComp,
    CaseIfTmpVarInForStmtComp,
    CaseInterfaceArrayNonStaticIndexComp,
    CaseLambdaConnectComp,
    CaseLambdaConnectWithListComp,
    CaseNestedIfComp,
    CaseNestedStructPackedArrayUpblkComp,
    CasePassThroughComp,
    CaseReducesInx3OutComp,
    CaseSequentialPassThroughComp,
    CaseSizeCastPaddingStructPort,
    CaseStructPackedArrayUpblkComp,
    CaseSVerilogReservedComp,
    NestedStructPackedPlusScalar,
    ThisIsABitStructWithSuperLongName,
    set_attributes,
)

CaseSizeCastPaddingStructPort = set_attributes( CaseSizeCastPaddingStructPort,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { { 32 { 1'b0 } }, in_ };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in___foo,
          output logic [63:0] out,
          input logic [0:0] reset
        );
          logic [31:0] in_;

          always_comb begin : upblk
            out = { { 32 { 1'b0 } }, in_ };
          end

          assign in_[31:0] = in___foo;

        endmodule
    '''
)

CaseLambdaConnectWithListComp = set_attributes( CaseLambdaConnectWithListComp,
    'REF_UPBLK',
    '''\
        always_comb begin : _lambda__s_out_1_
          out[1] = in_ + 32'd42;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [31:0] out__0,
          output logic [31:0] out__1,
          input logic [0:0] reset
        );
          logic [31:0] out [0:1];

          always_comb begin : _lambda__s_out_1_
            out[1] = in_ + 32'd42;
          end

          assign out__0 = out[0];
          assign out__1 = out[1];

        endmodule
    '''
)

CaseBits32x2ConcatFreeVarComp = set_attributes( CaseBits32x2ConcatFreeVarComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in_, 32'd0 };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in_,
          output logic [63:0] out,
          input logic [0:0] reset
        );

          always_comb begin : upblk
            out = { in_, 32'd0 };
          end

        endmodule
    '''
)

CaseBits32x2ConcatUnpackedSignalComp = set_attributes( CaseBits32x2ConcatUnpackedSignalComp,
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in___0,
          input logic [31:0] in___1,
          output logic [63:0] out,
          input logic [0:0] reset
        );
          logic [31:0] in_ [0:1];

          always_comb begin : upblk
            out = { in_[0], in_[1] };
          end

          assign in_[0] = in___0;
          assign in_[1] = in___1;

        endmodule
    '''
)

CaseConnectConstToOutComp = set_attributes( CaseConnectConstToOutComp,
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          output logic [31:0] out,
          input logic [0:0] reset
        );

          assign out = 32'd42;

        endmodule
    '''
)

CaseForRangeLowerUpperStepPassThroughComp = set_attributes( CaseForRangeLowerUpperStepPassThroughComp,
    'REF_UPBLK',
    '''\
        integer __loopvar__upblk_i;

        always_comb begin : upblk
          for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 2 )
            out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
          for ( __loopvar__upblk_i = 1; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 2 )
            out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in___0,
          input logic [31:0] in___1,
          input logic [31:0] in___2,
          input logic [31:0] in___3,
          input logic [31:0] in___4,
          output logic [31:0] out__0,
          output logic [31:0] out__1,
          output logic [31:0] out__2,
          output logic [31:0] out__3,
          output logic [31:0] out__4,
          input logic [0:0] reset
        );
          logic [31:0] in_ [0:4];
          logic [31:0] out [0:4];

          integer __loopvar__upblk_i;

          always_comb begin : upblk
            for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 2 )
              out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
            for ( __loopvar__upblk_i = 1; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 2 )
              out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
          end

          assign in_[0] = in___0;
          assign in_[1] = in___1;
          assign in_[2] = in___2;
          assign in_[3] = in___3;
          assign in_[4] = in___4;
          assign out__0 = out[0];
          assign out__1 = out[1];
          assign out__2 = out[2];
          assign out__3 = out[3];
          assign out__4 = out[4];

        endmodule
    '''
)

CaseIfExpInForStmtComp = set_attributes( CaseIfExpInForStmtComp,
    'REF_UPBLK',
    '''\
        integer __loopvar__upblk_i;

        always_comb begin : upblk
          for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
            out[__loopvar__upblk_i] = ( __loopvar__upblk_i == 1 ) ? in_[__loopvar__upblk_i] : in_[0];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in___0,
          input logic [31:0] in___1,
          input logic [31:0] in___2,
          input logic [31:0] in___3,
          input logic [31:0] in___4,
          output logic [31:0] out__0,
          output logic [31:0] out__1,
          output logic [31:0] out__2,
          output logic [31:0] out__3,
          output logic [31:0] out__4,
          input logic [0:0] reset
        );
          logic [31:0] in_ [0:4];
          logic [31:0] out [0:4];

          integer __loopvar__upblk_i;

          always_comb begin : upblk
            for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
              out[__loopvar__upblk_i] = ( __loopvar__upblk_i == 1 ) ? in_[__loopvar__upblk_i] : in_[0];
          end

          assign in_[0] = in___0;
          assign in_[1] = in___1;
          assign in_[2] = in___2;
          assign in_[3] = in___3;
          assign in_[4] = in___4;
          assign out__0 = out[0];
          assign out__1 = out[1];
          assign out__2 = out[2];
          assign out__3 = out[3];
          assign out__4 = out[4];

        endmodule
    '''
)

CaseIfExpUnaryOpInForStmtComp = set_attributes( CaseIfExpUnaryOpInForStmtComp,
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in___0,
          input logic [31:0] in___1,
          input logic [31:0] in___2,
          input logic [31:0] in___3,
          input logic [31:0] in___4,
          output logic [31:0] out__0,
          output logic [31:0] out__1,
          output logic [31:0] out__2,
          output logic [31:0] out__3,
          output logic [31:0] out__4,
          input logic [0:0] reset
        );
          logic [31:0] in_ [0:4];
          logic [31:0] out [0:4];

          integer __loopvar__upblk_i;

          always_comb begin : upblk
            for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
              out[__loopvar__upblk_i] = ( __loopvar__upblk_i == 1 ) ? ~in_[__loopvar__upblk_i] : in_[0];
          end

          assign in_[0] = in___0;
          assign in_[1] = in___1;
          assign in_[2] = in___2;
          assign in_[3] = in___3;
          assign in_[4] = in___4;
          assign out__0 = out[0];
          assign out__1 = out[1];
          assign out__2 = out[2];
          assign out__3 = out[3];
          assign out__4 = out[4];

        endmodule
    '''
)

CaseIfBoolOpInForStmtComp = set_attributes( CaseIfBoolOpInForStmtComp,
    'REF_UPBLK',
    '''\
        integer __loopvar__upblk_i;

        always_comb begin : upblk
          for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
            if ( ( in_[__loopvar__upblk_i] != 32'd0 ) && ( ( __loopvar__upblk_i < 4 ) ? in_[__loopvar__upblk_i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
              out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
            end
            else
              out[__loopvar__upblk_i] = 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in___0,
          input  logic [31:0]   in___1,
          input  logic [31:0]   in___2,
          input  logic [31:0]   in___3,
          input  logic [31:0]   in___4,
          output logic [31:0]   out__0,
          output logic [31:0]   out__1,
          output logic [31:0]   out__2,
          output logic [31:0]   out__3,
          output logic [31:0]   out__4,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in_ [0:4];
          logic [31:0]   out [0:4];

          integer __loopvar__upblk_i;

          always_comb begin : upblk
            for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
              if ( ( in_[__loopvar__upblk_i] != 32'd0 ) && ( ( __loopvar__upblk_i < 4 ) ? in_[__loopvar__upblk_i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
                out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
              end
              else
                out[__loopvar__upblk_i] = 32'd0;
          end

          // Connections
          assign in_[0] = in___0;
          assign in_[1] = in___1;
          assign in_[2] = in___2;
          assign in_[3] = in___3;
          assign in_[4] = in___4;
          assign out__0 = out[0];
          assign out__1 = out[1];
          assign out__2 = out[2];
          assign out__3 = out[3];
          assign out__4 = out[4];

        endmodule
    '''
)

CaseIfTmpVarInForStmtComp = set_attributes( CaseIfTmpVarInForStmtComp,
    'REF_UPBLK',
    '''\
        integer __loopvar__upblk_i;

        always_comb begin : upblk
          for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 ) begin
            if ( ( in_[__loopvar__upblk_i] != 32'd0 ) && ( ( __loopvar__upblk_i < 4 ) ? in_[__loopvar__upblk_i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
              __tmpvar__upblk_tmpvar = in_[__loopvar__upblk_i];
            end
            else
              __tmpvar__upblk_tmpvar = 32'd0;
            out[__loopvar__upblk_i] = __tmpvar__upblk_tmpvar;
          end
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in___0,
          input  logic [31:0]   in___1,
          input  logic [31:0]   in___2,
          input  logic [31:0]   in___3,
          input  logic [31:0]   in___4,
          output logic [31:0]   out__0,
          output logic [31:0]   out__1,
          output logic [31:0]   out__2,
          output logic [31:0]   out__3,
          output logic [31:0]   out__4,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in_ [0:4];
          logic [31:0]   out [0:4];

          // Temporary wire definitions
          logic [31:0]   __tmpvar__upblk_tmpvar;

          integer __loopvar__upblk_i;

          always_comb begin : upblk
            for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 ) begin
              if ( ( in_[__loopvar__upblk_i] != 32'd0 ) && ( ( __loopvar__upblk_i < 4 ) ? in_[__loopvar__upblk_i + 1] != 32'd0 : in_[4] != 32'd0 ) ) begin
                __tmpvar__upblk_tmpvar = in_[__loopvar__upblk_i];
              end
              else
                __tmpvar__upblk_tmpvar = 32'd0;
              out[__loopvar__upblk_i] = __tmpvar__upblk_tmpvar;
            end
          end

          // Connections
          assign in_[0] = in___0;
          assign in_[1] = in___1;
          assign in_[2] = in___2;
          assign in_[3] = in___3;
          assign in_[4] = in___4;
          assign out__0 = out[0];
          assign out__1 = out[1];
          assign out__2 = out[2];
          assign out__3 = out[3];
          assign out__4 = out[4];

        endmodule
    '''
)

CaseFixedSizeSliceComp = set_attributes( CaseFixedSizeSliceComp,
    'REF_UPBLK',
    '''\
        integer __loopvar__upblk_i;

        always_comb begin : upblk
          for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 2; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
            out[__loopvar__upblk_i] = in_[__loopvar__upblk_i * 8 +: 8];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [15:0]   in_,
          output logic [7:0]    out__0,
          output logic [7:0]    out__1,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [7:0]    out [0:1];

          integer __loopvar__upblk_i;

          always_comb begin : upblk
            for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 2; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
              out[__loopvar__upblk_i] = in_[__loopvar__upblk_i * 8 +: 8];
          end

          // Connections
          assign out__0 = out[0];
          assign out__1 = out[1];

        endmodule
    '''
)

CaseBits32FooInBits32OutComp = set_attributes( CaseBits32FooInBits32OutComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in___foo;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input logic [0:0] clk,
          input logic [31:0] in___foo,
          output logic [31:0] out,
          input logic [0:0] reset
        );
          logic [31:0] in_;

          always_comb begin : upblk
            out = in___foo;
          end

          assign in_[31:0] = in___foo;

        endmodule
    '''
)

CaseConstStructInstComp = set_attributes( CaseConstStructInstComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = 32'd0;
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );

          always_comb begin : upblk
            out = 32'd0;
          end

        endmodule
    '''
)

CaseStructPackedArrayUpblkComp = set_attributes( CaseStructPackedArrayUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in___foo[0], in___foo[1], in___foo[2] };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in___foo__0,
          input  logic [31:0]   in___foo__1,
          input  logic [31:0]   in___foo__2,
          input  logic [31:0]   in___foo__3,
          input  logic [31:0]   in___foo__4,
          output logic [95:0]   out,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in___foo [0:4];
          logic [159:0]  in_;

          always_comb begin : upblk
            out = { in___foo[0], in___foo[1], in___foo[2] };
          end

          // Connections
          assign in___foo[0] = in___foo__0;
          assign in___foo[1] = in___foo__1;
          assign in___foo[2] = in___foo__2;
          assign in___foo[3] = in___foo__3;
          assign in___foo[4] = in___foo__4;
          assign in_[159:128] = in___foo__4;
          assign in_[127:96] = in___foo__3;
          assign in_[95:64] = in___foo__2;
          assign in_[63:32] = in___foo__1;
          assign in_[31:0] = in___foo__0;

        endmodule
    '''
)

CaseNestedStructPackedArrayUpblkComp = set_attributes( CaseNestedStructPackedArrayUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = { in___bar[0], in___woo__foo, in___foo };
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in___foo,
          input  logic [31:0]   in___bar__0,
          input  logic [31:0]   in___bar__1,
          input  logic [31:0]   in___woo__foo,
          output logic [95:0]   out,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in___bar [0:1];
          logic [31:0]   in___woo;
          logic [127:0]  in_;

          always_comb begin : upblk
            out = { in___bar[0], in___woo__foo, in___foo };
          end

          // Connections
          assign in___bar[0] = in___bar__0;
          assign in___bar[1] = in___bar__1;
          assign in___woo[31:0] = in___woo__foo;
          assign in_[127:96] = in___foo;
          assign in_[95:64] = in___bar__1;
          assign in_[63:32] = in___bar__0;
          assign in_[31:0] = in___woo__foo;

        endmodule
    '''
)

CaseArrayBits32IfcInUpblkComp = set_attributes( CaseArrayBits32IfcInUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in___foo[1];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset,
          input  logic [31:0]   in___0__foo,
          input  logic [31:0]   in___1__foo,
          input  logic [31:0]   in___2__foo,
          input  logic [31:0]   in___3__foo,
          input  logic [31:0]   in___4__foo
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in___foo [0:4];

          always_comb begin : upblk
            out = in___foo[1];
          end

          // Connections
          assign in___foo[0] = in___0__foo;
          assign in___foo[1] = in___1__foo;
          assign in___foo[2] = in___2__foo;
          assign in___foo[3] = in___3__foo;
          assign in___foo[4] = in___4__foo;

        endmodule
    '''
)

CaseInterfaceArrayNonStaticIndexComp = set_attributes( CaseInterfaceArrayNonStaticIndexComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = in___foo[in___foo[0]];
        end
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset,
          input  logic [31:0]   in___0__foo,
          input  logic [31:0]   in___1__foo
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in___foo [0:1];

          always_comb begin : upblk
            out = in___foo[in___foo[0]];
          end

          // Connections
          assign in___foo[0] = in___0__foo;
          assign in___foo[1] = in___1__foo;

        endmodule
    '''
)

CaseBits32ArraySubCompAttrUpblkComp = set_attributes( CaseBits32ArraySubCompAttrUpblkComp,
    'REF_UPBLK',
    '''\
        always_comb begin : upblk
          out = b__out[1];
        end
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );

          // Connections
          assign out = 32'd42;

        endmodule

        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );
          // Struct/Array ports of sub-components in the form of wires
          logic [0:0]    b__clk [0:4];
          logic [31:0]   b__out [0:4];
          logic [0:0]    b__reset [0:4];

          // Sub-component declarations
          logic [0:0]    b__0__clk;
          logic [31:0]   b__0__out;
          logic [0:0]    b__0__reset;

          Bits32OutDrivenComp b__0
          (
            .clk            (         b__0__clk         ),
            .out            (         b__0__out         ),
            .reset          (        b__0__reset        )
          );

          logic [0:0]    b__1__clk;
          logic [31:0]   b__1__out;
          logic [0:0]    b__1__reset;

          Bits32OutDrivenComp b__1
          (
            .clk            (         b__1__clk         ),
            .out            (         b__1__out         ),
            .reset          (        b__1__reset        )
          );

          logic [0:0]    b__2__clk;
          logic [31:0]   b__2__out;
          logic [0:0]    b__2__reset;

          Bits32OutDrivenComp b__2
          (
            .clk            (         b__2__clk         ),
            .out            (         b__2__out         ),
            .reset          (        b__2__reset        )
          );

          logic [0:0]    b__3__clk;
          logic [31:0]   b__3__out;
          logic [0:0]    b__3__reset;

          Bits32OutDrivenComp b__3
          (
            .clk            (         b__3__clk         ),
            .out            (         b__3__out         ),
            .reset          (        b__3__reset        )
          );

          logic [0:0]    b__4__clk;
          logic [31:0]   b__4__out;
          logic [0:0]    b__4__reset;

          Bits32OutDrivenComp b__4
          (
            .clk            (         b__4__clk         ),
            .out            (         b__4__out         ),
            .reset          (        b__4__reset        )
          );

          // Connect struct/array ports and their wire forms
          assign b__0__clk = b__clk[0];
          assign b__1__clk = b__clk[1];
          assign b__2__clk = b__clk[2];
          assign b__3__clk = b__clk[3];
          assign b__4__clk = b__clk[4];
          assign b__out[0] = b__0__out;
          assign b__out[1] = b__1__out;
          assign b__out[2] = b__2__out;
          assign b__out[3] = b__3__out;
          assign b__out[4] = b__4__out;
          assign b__0__reset = b__reset[0];
          assign b__1__reset = b__reset[1];
          assign b__2__reset = b__reset[2];
          assign b__3__reset = b__reset[3];
          assign b__4__reset = b__reset[4];

          always_comb begin : upblk
            out = b__out[1];
          end

          // Connections
          assign b__clk[0] = clk;
          assign b__reset[0] = reset;
          assign b__clk[1] = clk;
          assign b__reset[1] = reset;
          assign b__clk[2] = clk;
          assign b__reset[2] = reset;
          assign b__clk[3] = clk;
          assign b__reset[3] = reset;
          assign b__clk[4] = clk;
          assign b__reset[4] = reset;

        endmodule
    '''
)

CaseConnectInToWireComp = set_attributes( CaseConnectInToWireComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        input  logic [31:0]   in___0,
        input  logic [31:0]   in___1,
        input  logic [31:0]   in___2,
        input  logic [31:0]   in___3,
        input  logic [31:0]   in___4,
        output logic [31:0]   out,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
        logic [31:0]   in_ [0:4];
    ''',
    'REF_PORTS_CONN',
    '''\
        assign in_[0] = in___0;
        assign in_[1] = in___1;
        assign in_[2] = in___2;
        assign in_[3] = in___3;
        assign in_[4] = in___4;
    ''',
    'REF_WIRE',
    '''\
        logic [31:0]   wire_ [0:4];
    ''',
    'REF_CONN',
    '''\
        assign out = wire_[2];
        assign wire_[0] = in_[0];
        assign wire_[1] = in_[1];
        assign wire_[2] = in_[2];
        assign wire_[3] = in_[3];
        assign wire_[4] = in_[4];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in___0,
          input  logic [31:0]   in___1,
          input  logic [31:0]   in___2,
          input  logic [31:0]   in___3,
          input  logic [31:0]   in___4,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in_ [0:4];

          // Wire declarations
          logic [31:0]   wire_ [0:4];

          // Connections
          assign in_[0] = in___0;
          assign in_[1] = in___1;
          assign in_[2] = in___2;
          assign in_[3] = in___3;
          assign in_[4] = in___4;
          assign out = wire_[2];
          assign wire_[0] = in_[0];
          assign wire_[1] = in_[1];
          assign wire_[2] = in_[2];
          assign wire_[3] = in_[3];
          assign wire_[4] = in_[4];

        endmodule
    '''
)

CaseConnectBitsConstToOutComp = set_attributes( CaseConnectBitsConstToOutComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        output logic [31:0]   out,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
    ''',
    'REF_PORTS_CONN',
    '''\
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = 32'd0;
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );

          // Connections
          assign out = 32'd0;

        endmodule
    '''
)

CaseConnectConstToOutComp = set_attributes( CaseConnectConstToOutComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        output logic [31:0]   out,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
    ''',
    'REF_PORTS_CONN',
    '''\
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = 32'd42;
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );

          // Connections
          assign out = 32'd42;

        endmodule
    '''
)

CaseConnectBitSelToOutComp = set_attributes( CaseConnectBitSelToOutComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        input  logic [31:0]   in_,
        output logic [0:0]   out,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
    ''',
    'REF_PORTS_CONN',
    '''\
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = in_[0:0];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in_,
          output logic [0:0]    out,
          input  logic [0:0]    reset
        );

          // Connections
          assign out = in_[0:0];

        endmodule
    '''
)

CaseConnectSliceToOutComp = set_attributes( CaseConnectSliceToOutComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        input  logic [31:0]   in_,
        output logic [3:0]   out,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
    ''',
    'REF_PORTS_CONN',
    '''\
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = in_[7:4];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in_,
          output logic [3:0]    out,
          input  logic [0:0]    reset
        );

          // Connections
          assign out = in_[7:4];

        endmodule
    '''
)

CaseConnectConstStructAttrToOutComp = set_attributes( CaseConnectConstStructAttrToOutComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        output logic [31:0]   out,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
    ''',
    'REF_PORTS_CONN',
    '''\
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = 32'd42;
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );

          // Connections
          assign out = 32'd42;

        endmodule
    '''
)

CaseConnectLiteralStructComp = set_attributes( CaseConnectLiteralStructComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        output logic [31:0]   out__foo,
        output logic [31:0]   out__bar__0,
        output logic [31:0]   out__bar__1,
        output logic [31:0]   out__woo__foo,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
        logic [31:0]   out__bar [0:1];
        logic [31:0]   out__woo;
        logic [127:0]  out;
    ''',
    'REF_PORTS_CONN',
    '''\
        assign out__bar__0 = out__bar[0];
        assign out__bar__1 = out__bar[1];
        assign out__woo__foo = out__woo[31:0];
        assign out__foo = out[127:96];
        assign out__bar__1 = out[95:64];
        assign out__bar__0 = out[63:32];
        assign out__woo__foo = out[31:0];
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = { 32'd42, { 32'd2, 32'd1 }, 32'd3 };
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out__foo,
          output logic [31:0]   out__bar__0,
          output logic [31:0]   out__bar__1,
          output logic [31:0]   out__woo__foo,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   out__bar [0:1];
          logic [31:0]   out__woo;
          logic [127:0]  out;

          // Connections
          assign out__bar__0 = out__bar[0];
          assign out__bar__1 = out__bar[1];
          assign out__woo__foo = out__woo[31:0];
          assign out__foo = out[127:96];
          assign out__bar__1 = out[95:64];
          assign out__bar__0 = out[63:32];
          assign out__woo__foo = out[31:0];
          assign out = { 32'd42, { 32'd2, 32'd1 }, 32'd3 };

        endmodule
    '''
)

CaseConnectArrayStructAttrToOutComp = set_attributes( CaseConnectArrayStructAttrToOutComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        input  logic [31:0]   in___foo__0,
        input  logic [31:0]   in___foo__1,
        input  logic [31:0]   in___foo__2,
        input  logic [31:0]   in___foo__3,
        input  logic [31:0]   in___foo__4,
        output logic [31:0]   out,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
        logic [31:0]   in___foo [0:4];
        logic [159:0]  in_;
    ''',
    'REF_PORTS_CONN',
    '''\
        assign in___foo[0] = in___foo__0;
        assign in___foo[1] = in___foo__1;
        assign in___foo[2] = in___foo__2;
        assign in___foo[3] = in___foo__3;
        assign in___foo[4] = in___foo__4;
        assign in_[159:128] = in___foo__4;
        assign in_[127:96] = in___foo__3;
        assign in_[95:64] = in___foo__2;
        assign in_[63:32] = in___foo__1;
        assign in_[31:0] = in___foo__0;
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out = in___foo[1];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in___foo__0,
          input  logic [31:0]   in___foo__1,
          input  logic [31:0]   in___foo__2,
          input  logic [31:0]   in___foo__3,
          input  logic [31:0]   in___foo__4,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in___foo [0:4];
          logic [159:0]  in_;

          // Connections
          assign in___foo[0] = in___foo__0;
          assign in___foo[1] = in___foo__1;
          assign in___foo[2] = in___foo__2;
          assign in___foo[3] = in___foo__3;
          assign in___foo[4] = in___foo__4;
          assign in_[159:128] = in___foo__4;
          assign in_[127:96] = in___foo__3;
          assign in_[95:64] = in___foo__2;
          assign in_[63:32] = in___foo__1;
          assign in_[31:0] = in___foo__0;
          assign out = in___foo[1];

        endmodule
    '''
)

CaseConnectNestedStructPackedArrayComp = set_attributes( CaseConnectNestedStructPackedArrayComp,
    'REF_PORTS_PORT',
    '''\
        input  logic [0:0]    clk,
        input  logic [31:0]   in___foo,
        input  logic [31:0]   in___bar__0,
        input  logic [31:0]   in___bar__1,
        input  logic [31:0]   in___woo__foo,
        output logic [95:0]   out,
        input  logic [0:0]    reset
    ''',
    'REF_PORTS_WIRE',
    '''\
        logic [31:0]   in___bar [0:1];
        logic [31:0]   in___woo;
        logic [127:0]  in_;
    ''',
    'REF_PORTS_CONN',
    '''\
        assign in___bar[0] = in___bar__0;
        assign in___bar[1] = in___bar__1;
        assign in___woo[31:0] = in___woo__foo;
        assign in_[127:96] = in___foo;
        assign in_[95:64] = in___bar__1;
        assign in_[63:32] = in___bar__0;
        assign in_[31:0] = in___woo__foo;
    ''',
    'REF_WIRE',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out[31:0] = in___foo;
        assign out[63:32] = in___woo__foo;
        assign out[95:64] = in___bar[0];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in___foo,
          input  logic [31:0]   in___bar__0,
          input  logic [31:0]   in___bar__1,
          input  logic [31:0]   in___woo__foo,
          output logic [95:0]   out,
          input  logic [0:0]    reset
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in___bar [0:1];
          logic [31:0]   in___woo;
          logic [127:0]  in_;

          // Connections
          assign in___bar[0] = in___bar__0;
          assign in___bar[1] = in___bar__1;
          assign in___woo[31:0] = in___woo__foo;
          assign in_[127:96] = in___foo;
          assign in_[95:64] = in___bar__1;
          assign in_[63:32] = in___bar__0;
          assign in_[31:0] = in___woo__foo;
          assign out[31:0] = in___foo;
          assign out[63:32] = in___woo__foo;
          assign out[95:64] = in___bar[0];

        endmodule
    '''
)

CaseConnectValRdyIfcComp = set_attributes( CaseConnectValRdyIfcComp,
    'REF_IFC_PORT',
    '''\
        input  logic [31:0]   in___msg,
        output logic [0:0]    in___rdy,
        input  logic [0:0]    in___val,
        output logic [31:0]   out__msg,
        input  logic [0:0]    out__rdy,
        output logic [0:0]    out__val
    ''',
    'REF_IFC_WIRE',
    '''\
    ''',
    'REF_IFC_CONN',
    '''\
    ''',
    'REF_CONN',
    '''\
        assign out__msg = in___msg;
        assign in___rdy = out__rdy;
        assign out__val = in___val;
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [0:0]    reset,
          input  logic [31:0]   in___msg,
          output logic [0:0]    in___rdy,
          input  logic [0:0]    in___val,
          output logic [31:0]   out__msg,
          input  logic [0:0]    out__rdy,
          output logic [0:0]    out__val
        );

          // Connections
          assign out__msg = in___msg;
          assign in___rdy = out__rdy;
          assign out__val = in___val;

        endmodule
    '''
)

CaseConnectArrayBits32FooIfcComp = set_attributes( CaseConnectArrayBits32FooIfcComp,
    'REF_IFC_PORT',
    '''\
        input  logic [31:0]   in___0__foo__foo,
        input  logic [31:0]   in___1__foo__foo,
        output logic [31:0]   out__0__foo__foo,
        output logic [31:0]   out__1__foo__foo
    ''',
    'REF_IFC_WIRE',
    '''\
        logic [31:0]   in___foo__foo [0:1];
        logic [31:0]   in___foo [0:1];
        logic [31:0]   out__foo__foo [0:1];
        logic [31:0]   out__foo [0:1];
    ''',
    'REF_IFC_CONN',
    '''\
        assign in___foo__foo[0] = in___0__foo__foo;
        assign in___foo__foo[1] = in___1__foo__foo;
        assign in___foo[0][31:0] = in___0__foo__foo;
        assign in___foo[1][31:0] = in___1__foo__foo;
        assign out__0__foo__foo = out__foo__foo[0];
        assign out__1__foo__foo = out__foo__foo[1];
        assign out__0__foo__foo = out__foo[0][31:0];
        assign out__1__foo__foo = out__foo[1][31:0];
    ''',
    'REF_CONN',
    '''\
        assign out__foo[0] = in___foo[0];
        assign out__foo[1] = in___foo[1];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [0:0]    reset,
          input  logic [31:0]   in___0__foo__foo,
          input  logic [31:0]   in___1__foo__foo,
          output logic [31:0]   out__0__foo__foo,
          output logic [31:0]   out__1__foo__foo
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   in___foo__foo [0:1];
          logic [31:0]   in___foo [0:1];
          logic [31:0]   out__foo__foo [0:1];
          logic [31:0]   out__foo [0:1];

          // Connections
          assign in___foo__foo[0] = in___0__foo__foo;
          assign in___foo__foo[1] = in___1__foo__foo;
          assign in___foo[0][31:0] = in___0__foo__foo;
          assign in___foo[1][31:0] = in___1__foo__foo;
          assign out__0__foo__foo = out__foo__foo[0];
          assign out__1__foo__foo = out__foo__foo[1];
          assign out__0__foo__foo = out__foo[0][31:0];
          assign out__1__foo__foo = out__foo[1][31:0];
          assign out__foo[0] = in___foo[0];
          assign out__foo[1] = in___foo[1];

        endmodule
    '''
)

CaseConnectArrayNestedIfcComp = set_attributes( CaseConnectArrayNestedIfcComp,
    'REF_IFC_PORT',
    '''\
        input  logic [0:0]    in___0__ctrl_foo,
        input  logic [0:0]    in___1__ctrl_foo,
        input  logic [31:0]   in___0__memifc__msg,
        input  logic [31:0]   in___1__memifc__msg,
        output logic [0:0]    in___0__memifc__rdy,
        output logic [0:0]    in___1__memifc__rdy,
        input  logic [0:0]    in___0__memifc__val,
        input  logic [0:0]    in___1__memifc__val,
        output logic [0:0]    out__0__ctrl_foo,
        output logic [0:0]    out__1__ctrl_foo,
        output logic [31:0]   out__0__memifc__msg,
        output logic [31:0]   out__1__memifc__msg,
        input  logic [0:0]    out__0__memifc__rdy,
        input  logic [0:0]    out__1__memifc__rdy,
        output logic [0:0]    out__0__memifc__val,
        output logic [0:0]    out__1__memifc__val
    ''',
    'REF_IFC_WIRE',
    '''\
        logic [0:0]    in___ctrl_foo [0:1];
        logic [31:0]   in___memifc__msg [0:1];
        logic [0:0]    in___memifc__rdy [0:1];
        logic [0:0]    in___memifc__val [0:1];
        logic [0:0]    out__ctrl_foo [0:1];
        logic [31:0]   out__memifc__msg [0:1];
        logic [0:0]    out__memifc__rdy [0:1];
        logic [0:0]    out__memifc__val [0:1];
    ''',
    'REF_IFC_CONN',
    '''\
        assign in___ctrl_foo[0] = in___0__ctrl_foo;
        assign in___ctrl_foo[1] = in___1__ctrl_foo;
        assign in___memifc__msg[0] = in___0__memifc__msg;
        assign in___memifc__msg[1] = in___1__memifc__msg;
        assign in___0__memifc__rdy = in___memifc__rdy[0];
        assign in___1__memifc__rdy = in___memifc__rdy[1];
        assign in___memifc__val[0] = in___0__memifc__val;
        assign in___memifc__val[1] = in___1__memifc__val;
        assign out__0__ctrl_foo = out__ctrl_foo[0];
        assign out__1__ctrl_foo = out__ctrl_foo[1];
        assign out__0__memifc__msg = out__memifc__msg[0];
        assign out__1__memifc__msg = out__memifc__msg[1];
        assign out__memifc__rdy[0] = out__0__memifc__rdy;
        assign out__memifc__rdy[1] = out__1__memifc__rdy;
        assign out__0__memifc__val = out__memifc__val[0];
        assign out__1__memifc__val = out__memifc__val[1];
    ''',
    'REF_CONN',
    '''\
        assign out__ctrl_foo[0] = in___ctrl_foo[0];
        assign out__memifc__msg[0] = in___memifc__msg[0];
        assign in___memifc__rdy[0] = out__memifc__rdy[0];
        assign out__memifc__val[0] = in___memifc__val[0];
        assign out__ctrl_foo[1] = in___ctrl_foo[1];
        assign out__memifc__msg[1] = in___memifc__msg[1];
        assign in___memifc__rdy[1] = out__memifc__rdy[1];
        assign out__memifc__val[1] = in___memifc__val[1];
    ''',
    'REF_SRC',
    '''\
        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [0:0]    reset,
          input  logic [0:0]    in___0__ctrl_foo,
          input  logic [0:0]    in___1__ctrl_foo,
          input  logic [31:0]   in___0__memifc__msg,
          input  logic [31:0]   in___1__memifc__msg,
          output logic [0:0]    in___0__memifc__rdy,
          output logic [0:0]    in___1__memifc__rdy,
          input  logic [0:0]    in___0__memifc__val,
          input  logic [0:0]    in___1__memifc__val,
          output logic [0:0]    out__0__ctrl_foo,
          output logic [0:0]    out__1__ctrl_foo,
          output logic [31:0]   out__0__memifc__msg,
          output logic [31:0]   out__1__memifc__msg,
          input  logic [0:0]    out__0__memifc__rdy,
          input  logic [0:0]    out__1__memifc__rdy,
          output logic [0:0]    out__0__memifc__val,
          output logic [0:0]    out__1__memifc__val
        );
          // Struct/Array ports in the form of wires
          logic [0:0]    in___ctrl_foo [0:1];
          logic [31:0]   in___memifc__msg [0:1];
          logic [0:0]    in___memifc__rdy [0:1];
          logic [0:0]    in___memifc__val [0:1];
          logic [0:0]    out__ctrl_foo [0:1];
          logic [31:0]   out__memifc__msg [0:1];
          logic [0:0]    out__memifc__rdy [0:1];
          logic [0:0]    out__memifc__val [0:1];

          // Connections
          assign in___ctrl_foo[0] = in___0__ctrl_foo;
          assign in___ctrl_foo[1] = in___1__ctrl_foo;
          assign in___memifc__msg[0] = in___0__memifc__msg;
          assign in___memifc__msg[1] = in___1__memifc__msg;
          assign in___0__memifc__rdy = in___memifc__rdy[0];
          assign in___1__memifc__rdy = in___memifc__rdy[1];
          assign in___memifc__val[0] = in___0__memifc__val;
          assign in___memifc__val[1] = in___1__memifc__val;
          assign out__0__ctrl_foo = out__ctrl_foo[0];
          assign out__1__ctrl_foo = out__ctrl_foo[1];
          assign out__0__memifc__msg = out__memifc__msg[0];
          assign out__1__memifc__msg = out__memifc__msg[1];
          assign out__memifc__rdy[0] = out__0__memifc__rdy;
          assign out__memifc__rdy[1] = out__1__memifc__rdy;
          assign out__0__memifc__val = out__memifc__val[0];
          assign out__1__memifc__val = out__memifc__val[1];
          assign out__ctrl_foo[0] = in___ctrl_foo[0];
          assign out__memifc__msg[0] = in___memifc__msg[0];
          assign in___memifc__rdy[0] = out__memifc__rdy[0];
          assign out__memifc__val[0] = in___memifc__val[0];
          assign out__ctrl_foo[1] = in___ctrl_foo[1];
          assign out__memifc__msg[1] = in___memifc__msg[1];
          assign in___memifc__rdy[1] = out__memifc__rdy[1];
          assign out__memifc__val[1] = in___memifc__val[1];

        endmodule
    '''
)

CaseBits32ConnectSubCompAttrComp = set_attributes( CaseBits32ConnectSubCompAttrComp,
    'REF_COMP_PORT',
    '''\
        logic [0:0]    b__clk;
        logic [31:0]   b__out;
        logic [0:0]    b__reset;

        Bits32OutDrivenComp b
        (
          .clk            (           b__clk          ),
          .out            (           b__out          ),
          .reset          (          b__reset         )
        );
    ''',
    'REF_COMP_WIRE',
    '''\
    ''',
    'REF_COMP_CONN',
    '''\
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );

          // Connections
          assign out = 32'd42;

        endmodule

        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );
          // Sub-component declarations
          logic [0:0]    b__clk;
          logic [31:0]   b__out;
          logic [0:0]    b__reset;

          Bits32OutDrivenComp b
          (
            .clk            (           b__clk          ),
            .out            (           b__out          ),
            .reset          (          b__reset         )
          );

          // Connections
          assign b__clk = clk;
          assign b__reset = reset;
          assign out = b__out;

        endmodule
    '''
)

CaseConnectArraySubCompArrayStructIfcComp = set_attributes( CaseConnectArraySubCompArrayStructIfcComp,
    'REF_COMP_PORT',
    '''\
        logic [0:0]    b__0__clk;
        logic [31:0]   b__0__out;
        logic [0:0]    b__0__reset;
        logic [31:0]   b__0__ifc__0__foo__0__foo;

        Bits32ArrayStructIfcComp b__0
        (
          .clk            (         b__0__clk         ),
          .out            (         b__0__out         ),
          .reset          (        b__0__reset        ),
          .ifc__0__foo__0__foo( b__0__ifc__0__foo__0__foo )
        );
    ''',
    'REF_COMP_WIRE',
    '''\
        logic [0:0]    b__clk [0:0];
        logic [31:0]   b__out [0:0];
        logic [0:0]    b__reset [0:0];
        logic [31:0]   b__ifc__foo__foo [0:0][0:0][0:0];
        logic [31:0]   b__ifc__foo [0:0][0:0][0:0];
          ''',
          'REF_COMP_CONN',
          '''\
        assign b__0__clk = b__clk[0];
        assign b__out[0] = b__0__out;
        assign b__0__reset = b__reset[0];
        assign b__0__ifc__0__foo__0__foo = b__ifc__foo__foo[0][0][0];
        assign b__0__ifc__0__foo__0__foo = b__ifc__foo[0][0][0][31:0];
    ''',
    'REF_SRC',
    '''\
        module Bits32ArrayStructIfcComp
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset,
          input  logic [31:0]   ifc__0__foo__0__foo
        );
          // Struct/Array ports in the form of wires
          logic [31:0]   ifc__foo__foo [0:0][0:0];
          logic [31:0]   ifc__foo [0:0][0:0];

          // Connections
          assign ifc__foo__foo[0][0] = ifc__0__foo__0__foo;
          assign ifc__foo[0][0][31:0] = ifc__0__foo__0__foo;
          assign out = ifc__foo__foo[0][0];

        endmodule

        module DUT
        (
          input  logic [0:0]    clk,
          input  logic [31:0]   in_,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );
          // Struct/Array ports of sub-components in the form of wires
          logic [0:0]    b__clk [0:0];
          logic [31:0]   b__out [0:0];
          logic [0:0]    b__reset [0:0];
          logic [31:0]   b__ifc__foo__foo [0:0][0:0][0:0];
          logic [31:0]   b__ifc__foo [0:0][0:0][0:0];

          // Sub-component declarations
          logic [0:0]    b__0__clk;
          logic [31:0]   b__0__out;
          logic [0:0]    b__0__reset;
          logic [31:0]   b__0__ifc__0__foo__0__foo;

          Bits32ArrayStructIfcComp b__0
          (
            .clk            (         b__0__clk         ),
            .out            (         b__0__out         ),
            .reset          (        b__0__reset        ),
            .ifc__0__foo__0__foo( b__0__ifc__0__foo__0__foo )
          );

          // Connect struct/array ports and their wire forms
          assign b__0__clk = b__clk[0];
          assign b__out[0] = b__0__out;
          assign b__0__reset = b__reset[0];
          assign b__0__ifc__0__foo__0__foo = b__ifc__foo__foo[0][0][0];
          assign b__0__ifc__0__foo__0__foo = b__ifc__foo[0][0][0][31:0];

          // Connections
          assign b__clk[0] = clk;
          assign b__reset[0] = reset;
          assign b__ifc__foo__foo[0][0][0] = in_;
          assign out = b__out[0];

        endmodule
    '''
)

CaseBits32ArrayConnectSubCompAttrComp = set_attributes( CaseBits32ArrayConnectSubCompAttrComp,
    'REF_COMP_PORT',
    '''\
        logic [0:0]    b__0__clk;
        logic [31:0]   b__0__out;
        logic [0:0]    b__0__reset;

        Bits32OutDrivenComp b__0
        (
          .clk            (         b__0__clk         ),
          .out            (         b__0__out         ),
          .reset          (        b__0__reset        )
        );

        logic [0:0]    b__1__clk;
        logic [31:0]   b__1__out;
        logic [0:0]    b__1__reset;

        Bits32OutDrivenComp b__1
        (
          .clk            (         b__1__clk         ),
          .out            (         b__1__out         ),
          .reset          (        b__1__reset        )
        );

        logic [0:0]    b__2__clk;
        logic [31:0]   b__2__out;
        logic [0:0]    b__2__reset;

        Bits32OutDrivenComp b__2
        (
          .clk            (         b__2__clk         ),
          .out            (         b__2__out         ),
          .reset          (        b__2__reset        )
        );

        logic [0:0]    b__3__clk;
        logic [31:0]   b__3__out;
        logic [0:0]    b__3__reset;

        Bits32OutDrivenComp b__3
        (
          .clk            (         b__3__clk         ),
          .out            (         b__3__out         ),
          .reset          (        b__3__reset        )
        );

        logic [0:0]    b__4__clk;
        logic [31:0]   b__4__out;
        logic [0:0]    b__4__reset;

        Bits32OutDrivenComp b__4
        (
          .clk            (         b__4__clk         ),
          .out            (         b__4__out         ),
          .reset          (        b__4__reset        )
        );
    ''',
    'REF_COMP_WIRE',
    '''\
        logic [0:0]    b__clk [0:4];
        logic [31:0]   b__out [0:4];
        logic [0:0]    b__reset [0:4];
    ''',
    'REF_COMP_CONN',
    '''\
        assign b__0__clk = b__clk[0];
        assign b__1__clk = b__clk[1];
        assign b__2__clk = b__clk[2];
        assign b__3__clk = b__clk[3];
        assign b__4__clk = b__clk[4];
        assign b__out[0] = b__0__out;
        assign b__out[1] = b__1__out;
        assign b__out[2] = b__2__out;
        assign b__out[3] = b__3__out;
        assign b__out[4] = b__4__out;
        assign b__0__reset = b__reset[0];
        assign b__1__reset = b__reset[1];
        assign b__2__reset = b__reset[2];
        assign b__3__reset = b__reset[3];
        assign b__4__reset = b__reset[4];
    ''',
    'REF_SRC',
    '''\
        module Bits32OutDrivenComp
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );

          // Connections
          assign out = 32'd42;

        endmodule

        module DUT
        (
          input  logic [0:0]    clk,
          output logic [31:0]   out,
          input  logic [0:0]    reset
        );
          // Struct/Array ports of sub-components in the form of wires
          logic [0:0]    b__clk [0:4];
          logic [31:0]   b__out [0:4];
          logic [0:0]    b__reset [0:4];

          // Sub-component declarations
          logic [0:0]    b__0__clk;
          logic [31:0]   b__0__out;
          logic [0:0]    b__0__reset;

          Bits32OutDrivenComp b__0
          (
            .clk            (         b__0__clk         ),
            .out            (         b__0__out         ),
            .reset          (        b__0__reset        )
          );

          logic [0:0]    b__1__clk;
          logic [31:0]   b__1__out;
          logic [0:0]    b__1__reset;

          Bits32OutDrivenComp b__1
          (
            .clk            (         b__1__clk         ),
            .out            (         b__1__out         ),
            .reset          (        b__1__reset        )
          );

          logic [0:0]    b__2__clk;
          logic [31:0]   b__2__out;
          logic [0:0]    b__2__reset;

          Bits32OutDrivenComp b__2
          (
            .clk            (         b__2__clk         ),
            .out            (         b__2__out         ),
            .reset          (        b__2__reset        )
          );

          logic [0:0]    b__3__clk;
          logic [31:0]   b__3__out;
          logic [0:0]    b__3__reset;

          Bits32OutDrivenComp b__3
          (
            .clk            (         b__3__clk         ),
            .out            (         b__3__out         ),
            .reset          (        b__3__reset        )
          );

          logic [0:0]    b__4__clk;
          logic [31:0]   b__4__out;
          logic [0:0]    b__4__reset;

          Bits32OutDrivenComp b__4
          (
            .clk            (         b__4__clk         ),
            .out            (         b__4__out         ),
            .reset          (        b__4__reset        )
          );

          // Connect struct/array ports and their wire forms
          assign b__0__clk = b__clk[0];
          assign b__1__clk = b__clk[1];
          assign b__2__clk = b__clk[2];
          assign b__3__clk = b__clk[3];
          assign b__4__clk = b__clk[4];
          assign b__out[0] = b__0__out;
          assign b__out[1] = b__1__out;
          assign b__out[2] = b__2__out;
          assign b__out[3] = b__3__out;
          assign b__out[4] = b__4__out;
          assign b__0__reset = b__reset[0];
          assign b__1__reset = b__reset[1];
          assign b__2__reset = b__reset[2];
          assign b__3__reset = b__reset[3];
          assign b__4__reset = b__reset[4];

          // Connections
          assign b__clk[0] = clk;
          assign b__reset[0] = reset;
          assign b__clk[1] = clk;
          assign b__reset[1] = reset;
          assign b__clk[2] = clk;
          assign b__reset[2] = reset;
          assign b__clk[3] = clk;
          assign b__reset[3] = reset;
          assign b__clk[4] = clk;
          assign b__reset[4] = reset;
          assign out = b__out[1];

        endmodule
    '''
)



"""
========================================================================
Tests for CL queues
========================================================================

Author: Yanghui Ou
  Date: Mar 24, 2019
"""
from itertools import product

import pytest

from pymtl3 import *
from pymtl3.stdlib.test import TestVectorSimulator
from pymtl3.stdlib.test.test_sinks import TestSinkRTL
from pymtl3.stdlib.test.test_srcs import TestSrcCL

from .queues import (
    BypassQueue1EntryRTL,
    BypassQueueRTL,
    NormalQueue1EntryRTL,
    NormalQueueRTL,
    PipeQueue1EntryRTL,
    PipeQueueRTL,
)

#-------------------------------------------------------------------------
# TestVectorSimulator test
#-------------------------------------------------------------------------

def run_tv_test( dut, test_vectors ):

  # Define input/output functions

  def tv_in( dut, tv ):
    dut.enq.en  = tv[0]
    dut.enq.msg = tv[2]
    dut.deq.en  = tv[3]

  def tv_out( dut, tv ):
    if tv[1] != '?': assert dut.enq.rdy == tv[1]
    if tv[4] != '?': assert dut.deq.rdy == tv[4]
    if tv[5] != '?': assert dut.deq.ret == tv[5]

  # Run the test

  sim = TestVectorSimulator( dut, test_vectors, tv_in, tv_out )
  sim.run_test()

def test_pipe_Bits():

  B1  = mk_bits(1)
  B32 = mk_bits(32)
  run_tv_test( NormalQueueRTL( Bits32, 2 ), [
    #  enq.en  enq.rdy enq.msg   deq.en  deq.rdy deq.ret
    [  B1(1),  B1(1),  B32(123), B1(0),  B1(0),    '?'    ],
    [  B1(1),  B1(1),  B32(345), B1(0),  B1(1),  B32(123) ],
    [  B1(0),  B1(0),  B32(567), B1(0),  B1(1),  B32(123) ],
    [  B1(0),  B1(0),  B32(567), B1(1),  B1(1),  B32(123) ],
    [  B1(0),  B1(1),  B32(567), B1(1),  B1(1),  B32(345) ],
    [  B1(1),  B1(1),  B32(567), B1(0),  B1(0),    '?'    ],
    [  B1(1),  B1(1),  B32(0  ), B1(1),  B1(1),  B32(567) ],
    [  B1(1),  B1(1),  B32(1  ), B1(1),  B1(1),  B32(0  ) ],
    [  B1(1),  B1(1),  B32(2  ), B1(1),  B1(1),  B32(1  ) ],
    [  B1(0),  B1(1),  B32(2  ), B1(1),  B1(1),  B32(2  ) ],
] )

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, QType, src_msgs, sink_msgs ):

    s.src  = TestSrcCL ( MsgType, src_msgs )
    s.dut  = QType( MsgType )
    s.sink = TestSinkRTL( MsgType, sink_msgs )

    connect( s.src.send, s.dut.enq   )
    connect( s.dut.deq,  s.sink.recv )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{} ({}) {}".format(
      s.src.line_trace(), s.dut.line_trace(), s.sink.line_trace() )

#-------------------------------------------------------------------------
# run_sim
#-------------------------------------------------------------------------

def run_sim( th, max_cycles=100 ):

  # Create a simulator

  th.apply( SimulationPass() )
  th.sim_reset()

  print("")
  ncycles = 0
  print("{:2}:{}".format( ncycles, th.line_trace() ))
  while not th.done() and ncycles < max_cycles:
    th.tick()
    ncycles += 1
    print("{:2}:{}".format( ncycles, th.line_trace() ))

  # Check timeout

  assert ncycles < max_cycles

  th.tick()
  th.tick()
  th.tick()

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

test_msgs = [ Bits16( 4 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ) ]

arrival_normal = [ 2, 4, 6, 8 ]
arrival_pipe   = [ 2, 3, 4, 5 ]
arrival_bypass = [ 1, 2, 3, 4 ]


def test_normal1_simple():
  th = TestHarness( Bits16, NormalQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_normal )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th )

def test_normal2_simple():
  th = TestHarness( Bits16, NormalQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_pipe )
  th.set_param( "top.dut.construct", num_entries = 2 )
  run_sim( th )

def test_pipe1_simple():
  th = TestHarness( Bits16, PipeQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_pipe )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th )

def test_pipe1_backpressure():
  th = TestHarness( Bits16, PipeQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th )

def test_pipe2_backpressure():
  th = TestHarness( Bits16, PipeQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 2 )
  run_sim( th )

def test_bypass1_simple():
  th = TestHarness( Bits16, BypassQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", arrival_time = arrival_bypass )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th )

def test_bypass1_backpressure():
  th = TestHarness( Bits16, BypassQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 1 )
  run_sim( th )

def test_bypass2_sparse():
  th = TestHarness( Bits16, BypassQueueRTL, test_msgs, test_msgs )
  th.set_param( "top.src.construct", interval_delay = 3 )
  th.set_param( "top.dut.construct", num_entries = 2 )
  run_sim( th )

@pytest.mark.parametrize(
  'QType, num_entries',
  product( [ NormalQueueRTL, PipeQueueRTL, BypassQueueRTL ], [ 8, 10, 12, 16 ] )
)
def test_large_backpressure( QType, num_entries ):
  msgs = test_msgs * 8
  th = TestHarness( Bits16, QType, msgs, msgs )
  th.set_param( "top.sink.construct", initial_delay = 20 )
  th.set_param( "top.dut.construct", num_entries = 16 )
  run_sim( th )

@pytest.mark.parametrize(
  'QType', [ NormalQueue1EntryRTL, PipeQueue1EntryRTL, BypassQueue1EntryRTL ]
)
def test_single_simple( QType ):
  th = TestHarness( Bits16, QType, test_msgs, test_msgs )
  run_sim( th )

@pytest.mark.parametrize(
  'QType', [ NormalQueue1EntryRTL, PipeQueue1EntryRTL, BypassQueue1EntryRTL ]
)
def test_single_backpressure( QType ):
  th = TestHarness( Bits16, QType, test_msgs, test_msgs )
  th.set_param( "top.sink.construct", initial_delay = 10, interval_delay=2 )
  run_sim( th )



#=========================================================================
# RTLIRTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 15, 2019
"""Provide translators that convert RTLIR to backend representation."""

from .BaseRTLIRTranslator import TranslatorMetadata
from .behavioral import BehavioralTranslator
from .errors import RTLIRTranslationError
from .structural import StructuralTranslator


def mk_RTLIRTranslator( _StructuralTranslator, _BehavioralTranslator ):
  """Return an RTLIRTranslator from the two given translators."""
  class _RTLIRTranslator( _StructuralTranslator, _BehavioralTranslator ):
    """Template translator class parameterized by behavioral/structural translator.

    Components are assembled here because they have both behavioral
    and structural parts. This translator also generates the overall
    code layout of the backend representation.
    """

    # Override
    def __init__( s, top ):
      super().__init__( top )

    # Override
    def clear( s, tr_top ):
      super().clear( tr_top )
      s.hierarchy = TranslatorMetadata()

    def _gen_hierarchy_metadata( s, structural_ns, hierarchy_ns ):
      metadata = getattr( s.structural, structural_ns, [] )
      List = getattr( s.hierarchy, hierarchy_ns )
      for Type, data in metadata:
        if not any( x[0] == Type for x in List ):
          List.append( ( Type, data ) )

    # Override
    def translate( s, tr_top ):

      def get_component_nspace( namespace, m ):
        ns = TranslatorMetadata()
        for name, metadata_d in vars(namespace).items():
          # Hierarchical metadata will not be added
          if m in metadata_d:
            setattr( ns, name, metadata_d[m] )
        return ns

      def translate_component( m, components, translated ):
        for child in sorted(m.get_child_components(), key = lambda x: x._dsl.my_name):
          translate_component( child, components, translated )
        if not s.structural.component_unique_name[m] in translated:
          components.append(
            s.rtlir_tr_component(
              get_component_nspace( s.behavioral, m ),
              get_component_nspace( s.structural, m ),
          ) )
          translated.append( s.structural.component_unique_name[m] )
        s._gen_hierarchy_metadata( 'decl_type_vector', 'decl_type_vector' )
        s._gen_hierarchy_metadata( 'decl_type_array', 'decl_type_array'   )
        s._gen_hierarchy_metadata( 'decl_type_struct', 'decl_type_struct' )

      # Clear all translator metadata
      s.clear( tr_top )

      s.component = {}
      # Generate backend representation for each component
      s.hierarchy.components = []
      s.hierarchy.decl_type_vector = []
      s.hierarchy.decl_type_array = []
      s.hierarchy.decl_type_struct = []

      try:
        s.translate_behavioral( s.tr_top )
        s.translate_structural( s.tr_top )
        translate_component( s.tr_top, s.hierarchy.components, [] )
      except AssertionError as e:
        msg = '' if e.args[0] is None else e.args[0]
        raise RTLIRTranslationError( s.tr_top, msg )

      # Generate the representation for all components
      s.hierarchy.component_src = s.rtlir_tr_components(s.hierarchy.components)

      # Generate the final backend code layout
      s.hierarchy.src = s.rtlir_tr_src_layout( s.hierarchy )

    #---------------------------------------------------------------------
    # Methods to be implemented by the backend translator
    #---------------------------------------------------------------------

    def rtlir_tr_src_layout( s, hierarchy ):
      raise NotImplementedError()

    def rtlir_tr_components( s, components ):
      raise NotImplementedError()

    def rtlir_tr_component( s, behavioral, structural ):
      raise NotImplementedError()

  return _RTLIRTranslator

RTLIRTranslator = mk_RTLIRTranslator( BehavioralTranslator, StructuralTranslator )



"""
========================================================================
Encoder.py
========================================================================
A priority encoder for arbitration

Author : Yanghui Ou, Cheng Tan
  Date : Mar 1, 2019
"""
from pymtl3 import *


class Encoder( Component ):
  def construct( s, in_nbits, out_nbits ):

    # Interface

    InType  = mk_bits(in_nbits)
    OutType = mk_bits(out_nbits)

    s.in_ =  InPort( InType )
    s.out = OutPort( OutType )

    # Constants

    s.in_nbits  = in_nbits
    s.out_nbits = out_nbits

    # Logic

    @s.update
    def encode():
      s.out = OutType( 0 )
      for i in range( s.in_nbits ):
        if s.in_[i]:
          s.out = OutType( i )

  def line_trace( s ):
    return "in:{:0>{n}b} | out:{}".format(
      int( s.in_ ), s.out, n=s.in_nbits
    )



#=========================================================================
# YosysTranslator_L2_cases_test.py
#=========================================================================
"""Test the yosys-SystemVerilog translator."""

import pytest

from pymtl3.passes.backends.sverilog.util.test_utility import check_eq
from pymtl3.passes.rtlir.util.test_utility import get_parameter

from ...testcases import (
    CaseConnectPassThroughLongNameComp,
    ThisIsABitStructWithSuperLongName,
)
from ..behavioral.test.YosysBehavioralTranslatorL2_test import test_yosys_behavioral_L2
from ..behavioral.test.YosysBehavioralTranslatorL3_test import test_yosys_behavioral_L3
from ..structural.test.YosysStructuralTranslatorL2_test import test_yosys_structural_L2
from ..YosysTranslator import YosysTranslator


def run_test( case, m ):
  m.elaborate()
  tr = YosysTranslator( m )
  tr.translate( m )
  check_eq( tr.hierarchy.src, case.REF_SRC )

@pytest.mark.parametrize(
  'case', get_parameter('case', test_yosys_behavioral_L2) + \
          get_parameter('case', test_yosys_behavioral_L3) + \
          get_parameter('case', test_yosys_structural_L2)
)
def test_yosys_L2( case ):
  run_test( case, case.DUT() )

def test_long_component_name():
  args = [ThisIsABitStructWithSuperLongName]*7
  run_test( CaseConnectPassThroughLongNameComp, CaseConnectPassThroughLongNameComp.DUT(*args) )

@pytest.mark.xfail(run=False, reason="TODO: resolving BitStructs according to name AND fields")
def test_struct_uniqueness():
  class A:
    @bitstruct
    class ST:
      a_foo: Bits16
      a_bar: Bits32

  class B:
    @bitstruct
    class ST:
      b_foo: Bits16
      b_bar: Bits32

  @bitstruct
  class COMB:
    fst: A.ST
    snd: B.ST

  class Top( Component ):
    def construct( s ):
      s.out = OutPort( COMB )
      connect( s.out, COMB(A.ST(1, 2), B.ST(3, 4)) )
  a = Top()
  a.REF_SRC = \
"""
typedef struct packed {
  logic [15:0] foo;
  logic [31:0] bar;
} ST;

typedef struct packed {
  ST fst;
  ST snd;
} COMB;

module Top
(
  input logic [0:0] clk,
  output COMB out,
  input logic [0:0] reset
);

  assign out = { { 16'd1, 32'd2 }, { 16'd3, 32'd4 } };

endmodule
"""
  a.YOSYS_REF_SRC = \
"""
module Top
(
  input logic [0:0] clk,
  output logic [15:0] out__fst__foo,
  output logic [31:0] out__fst__bar,
  output logic [15:0] out__snd__foo,
  output logic [31:0] out__snd__bar,
  input logic [0:0] reset
);
  logic [47:0]  out__fst;
  logic [47:0]  out__snd;
  logic [95:0]  out;

  assign out__fst__foo = out__fst[47:32];
  assign out__fst__bar = out__fst[31:0];
  assign out__snd__foo = out__snd[47:32];
  assign out__snd__bar = out__snd[31:0];
  assign out__fst__foo = out[95:80];
  assign out__fst__bar = out[79:48];
  assign out__snd__foo = out[47:32];
  assign out__snd__bar = out[31:0];
  assign out = { { 16'd1, 32'd2 }, { 16'd3, 32'd4 } };

endmodule
"""
  run_test( a, Top() )



#=========================================================================
# StructuralRTLIRGenL2Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 19, 2019
"""Test the generation of level 2 structural RTLIR."""

from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL2Pass import (
    StructuralRTLIRGenL2Pass,
)
from pymtl3.passes.rtlir.structural.StructuralRTLIRSignalExpr import *
from pymtl3.passes.testcases import (
    CaseConnectArrayStructAttrToOutComp,
    CaseConnectStructAttrToOutComp,
)

from .StructuralRTLIRGenL1Pass_test import gen_connections


def test_L2_struct_attr():
  a = CaseConnectStructAttrToOutComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL2Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  assert ns.connections == \
    [(StructAttr(CurCompAttr(comp, 'in_'), 'foo'), CurCompAttr(comp, 'out'))]

def test_L2_packed_index():
  a = CaseConnectArrayStructAttrToOutComp.DUT()
  a.elaborate()
  a.apply( StructuralRTLIRGenL2Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  assert ns.connections == \
    [(PackedIndex(StructAttr(CurCompAttr(comp, 'in_'), 'foo'), 1),
      CurCompAttr(comp, 'out'))]



"""
==========================================================================
bitstruct.py
==========================================================================
APIs to generate a bitstruct type. Using decorators and type annotations
to create bit struct is much inspired by python3 dataclass implementation.
Note that the implementation (such as the _CAPITAL constants to add some
private metadata) in this file is very similar to the **original python3
dataclass implementation**. The syntax of creating bit struct is very
similar to that of python3 dataclass.
https://github.com/python/cpython/blob/master/Lib/dataclasses.py

For example,

@bitstruct
class Point:
  x : Bits4
  y : Bits4

will automatically generate some methods, such as __init__, __str__,
__repr__, for the Point class.

Similar to the built-in dataclasses module, we also provide a
mk_bitstruct function for user to dynamically generate bit struct types.
For example,

mk_bitstruct( 'Pixel',{
    'r' : Bits4,
    'g' : Bits4,
    'b' : Bits4,
  },
  name_space = {
    '__str__' : lambda self: f'({self.r},{self.g},{self.b})'
  }
)

is equivalent to:

@bitstruct
class Pixel:
  r : Bits4
  g : Bits4
  b : Bits4

  def __str__( self ):
    return f'({self.r},{self.g},{self.b})'

Author : Yanghui Ou, Shunning Jiang
  Date : Oct 19, 2019
"""
import keyword
import types
import warnings

import py

from .bits_import import *

#-------------------------------------------------------------------------
# Constants
#-------------------------------------------------------------------------

# Object with this attribute is considered as bit struct, as we assume
# only the bitstruct decorator will stamp this attribute to a class. This
# attribute also stores the field information and can be used for
# translation.
#
# The original dataclass use hasattr( cls, _FIELDS ) to check dataclass.
# We do this here as well

_FIELDS = '__bitstruct_fields__'

def is_bitstruct_inst( obj ):
  """Returns True if obj is an instance of a dataclass."""
  return hasattr(type(obj), _FIELDS)

def is_bitstruct_class(cls):
  """Returns True if obj is a dataclass ."""
  return isinstance(cls, type) and hasattr(cls, _FIELDS)

_DEFAULT_SELF_NAME = 's'
_ANTI_CONFLICT_SELF_NAME = '__bitstruct_self__'

#-------------------------------------------------------------------------
# _create_fn
#-------------------------------------------------------------------------
# A helper function that creates a function based on
# - fn_name : name of the function
# - args_lst : a list of arguments in string
# - body_lst : a list of statement of the function body in string

# Also note that this whole _create_fn thing is similar to the original
# dataclass implementation!

def _create_fn( fn_name, args_lst, body_lst, _globals=None, class_method=False ):
  # Assemble argument string and body string
  args = ', '.join(args_lst)
  body = '\n'.join(f'  {statement}' for statement in body_lst)

  # Assemble the source code and execute it
  src = '@classmethod\n' if class_method else ''
  src += f'def {fn_name}({args}):\n{body}'
  _locals = {}
  exec( py.code.Source(src).compile(), _globals, _locals )
  return _locals[fn_name]

#-------------------------------------------------------------------------
# _mk_init_arg
#-------------------------------------------------------------------------
# Creates a init argument string from a field.
#
# Shunning: I revamped the whole thing because they are indeed mutable
# objects.

def _mk_init_arg( name, type_ ):
  # default is always None
  if isinstance( type_, list ) or is_bitstruct_class( type_ ):
    return f'{name} = None'
  return f'{name} = 0'

#-------------------------------------------------------------------------
# _mk_init_body
#-------------------------------------------------------------------------
# Creates one line of __init__ body from a field
# to globals.

def _mk_init_body( self_name, name, type_ ):
  def _recursive_generate_init( x ):
    if isinstance( x, list ):
      return f"[{', '.join( [ _recursive_generate_init(x[0]) ] * len(x) )}]"
    return f"_type_{name}()"

  if isinstance( type_, list ) or is_bitstruct_class( type_ ):
    return f'{self_name}.{name} = {name} or {_recursive_generate_init(type_)}'

  assert issubclass( type_, Bits )
  return f'{self_name}.{name} = _type_{name}({name})'

#-------------------------------------------------------------------------
# _mk_tuple_str
#-------------------------------------------------------------------------
# Creates a tuple of string representations of each field. For example,
# if the self_name is 'self' and fields is [ 'x', 'y' ], it will return
# ('self.x', 'self.y'). This is used for creating the default __eq__ and
# __hash__ function.

def _mk_tuple_str( self_name, fields ):
  return f'({",".join([f"{self_name}.{name}" for name in fields])},)'

#-------------------------------------------------------------------------
# _mk_init_fn
#-------------------------------------------------------------------------
# Creates a __init__ function based on fields. For example, if fields
# contains two field x (Bits4) and y (Bits4), _mk_init_fn will return a
# function that looks like the following:
#
# def __init__( s, x = 0, y = 0, z = None, p = None ):
#   s.x = _type_x(x)
#   s.y = _type_y(y)
#   s.z = z or _type_z()
#   s.p = p or [ _type_p(), _type_p() ]
#
# NOTE:
# _mk_init_fn also takes as argument the name of self in case there is a
# field with name 's' or 'self'.
#
# TODO: should we provide a __post_init__ function like dataclass does?

def _mk_init_fn( self_name, fields ):

  # Register necessary types in _globals
  _globals = {}

  for name, type_ in fields.items():
    if isinstance( type_, list ):
      x = type_[0]
      while isinstance( x, list ):
        x = x[0]
      _globals[ f"_type_{name}" ] = x
    else:
      assert issubclass( type_, Bits ) or is_bitstruct_class( type_ )
      _globals[ f"_type_{name}" ] = type_

  return _create_fn(
    '__init__',
    [ self_name ] + [ _mk_init_arg( *field ) for field in fields.items() ],
    [ _mk_init_body( self_name, *field ) for field in fields.items() ],
    _globals = _globals,
  )

#-------------------------------------------------------------------------
# _mk_str_fn
#-------------------------------------------------------------------------
# Creates a __str__ function based on fields. For example, if fields
# contains two field x (Bits4) and y (Bits4), _mk_str_fn will return a
# function that looks like the following:
#
# def __str__( self ):
#   return f'{self.x}:{self.y}'

def _mk_str_fn( fields ):
  return _create_fn(
    '__str__',
    [ 'self' ],
    [ 'return f"' +
      ':'.join([ f'{{self.{name}}}' for name in fields ]) + '"']
  )

#-------------------------------------------------------------------------
# _mk_repr_fn
#-------------------------------------------------------------------------
# Creates a __repr__ function based on fields. For example, if fields
# contains two field x (Bits4) and y (Bits4), _mk_repr_fn will return a
# function that looks like the following:
#
# def __repr__( self ):
#  return self.__class__.__name__ + f'(x={self.x!r}, y={self.y!r})'

def _mk_repr_fn( fields ):
  return _create_fn(
    '__repr__',
    [ 'self' ],
    [ 'return self.__class__.__name__ + f"(' +
      ','.join([ f'{{self.{name}!r}}' for name in fields ]) +
      ')"']
  )

#-------------------------------------------------------------------------
# _mk_eq_fn
#-------------------------------------------------------------------------
# Creates a __eq__ function based on fields. By default it just compares
# each field. For example, if fields contains two field x (Bits4) and y
# (Bits4), _mk_eq_fn will return a function that looks like the
# following:
#
# def __eq__( self, other ):
#   if other.__class__ is self.__class__:
#     return (self.x,self.y,) == (other.x,other.y,)
#   else:
#     raise NotImplemented

def _mk_eq_fn( fields ):
  self_tuple  = _mk_tuple_str( 'self', fields )
  other_tuple = _mk_tuple_str( 'other', fields )
  return _create_fn(
    '__eq__',
    [ 'self', 'other' ],
    [ f'return (other.__class__ is self.__class__) and {self_tuple} == {other_tuple}' ]
  )

#-------------------------------------------------------------------------
# _mk_hash_fn
#-------------------------------------------------------------------------
# Creates a __hash__ function based on fields. By default it just hashes
# all fields. For example, if fields contains two field x (Bits4) and y
# (Bits4), _mk_hash_fn will return a function that looks like the
# following:
#
# def __hash__( self ):
#   return hash((self.x,self.y,))

def _mk_hash_fn( fields ):
  self_tuple = _mk_tuple_str( 'self', fields )
  return _create_fn(
    '__hash__',
    [ 'self' ],
    [ f'return hash({self_tuple})' ]
  )

#-------------------------------------------------------------------------
# _mk_ff_fn
#-------------------------------------------------------------------------
# Creates __ilshift__ and _flip functions that looks like the following:
#
# def __ilshift__( self, other ):
#   self.x <<= other.x
#   for i in range(5):
#     for j in range(6):
#       self.y[i][j] <<= other.y[i][j]
#
# def _flip( self, other ):
#   self.x._flip()
#   for i in range(5):
#     for j in range(6):
#       self.y[i][j]._flip()

def _mk_ff_fn( fields ):
  ilshift_strs = []
  flip_strs    = []
  for name, type_ in fields.items():
    if isinstance( type_, list ):
      i = 0
      loop = f"{' '*i}for i{i} in range({len(type_)}):"
      ilshift_strs.append(loop)
      flip_strs   .append(loop)
      type_ = type_[0]
      i = 1
      while isinstance( type_, list ):
        loop = f"{' '*(i*2)}for i{i} in range({len(type_)}):"
        ilshift_strs.append(loop)
        flip_strs   .append(loop)
        type_ = type_[0]
        i += 1

      indices = ''.join( [ f'[i{k}]' for k in range(i)] )
      ilshift_strs.append( f"{' '*(i*2)}self.{name}{indices} <<= o.{name}{indices}" )
      flip_strs   .append( f"{' '*(i*2)}self.{name}{indices}._flip()" )

    else:
      ilshift_strs.append( f'self.{name} <<= o.{name}' )
      flip_strs.append( f'self.{name}._flip()' )

  return _create_fn(
    '__ilshift__',
    [ 'self', 'o' ],
    ilshift_strs + [ 'return self' ],
  ), _create_fn(
    '_flip',
    [ 'self' ],
    flip_strs,
  ),

#-------------------------------------------------------------------------
# _mk_clone_fn
#-------------------------------------------------------------------------
# Creates clone function that looks like the following:
# Use this clone function in any place that you need to perform a
# deepcopy on a bitstruct.
#
# def clone( self ):
#   return self.__class__( self.x.clone(), [ self.y[0].clone(), self.y[1].clone() ]  )

def _gen_list_clone_strs( type_, prefix='' ):
  if isinstance( type_, list ):
    return "[" + ",".join( [ _gen_list_clone_strs( type_[0], f"{prefix}[{i}]" )
                        for i in range(len(type_)) ] ) + "]"
  else:
    return f"{prefix}.clone()"

def _mk_clone_fn( fields ):
  clone_strs = [ 'return self.__class__(' ]

  for name, type_ in fields.items():
    clone_strs.append( "  " + _gen_list_clone_strs( type_, f'self.{name}' ) + "," )

  return _create_fn(
    'clone',
    [ 'self' ],
    clone_strs + [ ')' ],
  )

def _mk_deepcopy_fn( fields ):
  clone_strs = [ 'return self.__class__(' ]

  for name, type_ in fields.items():
    clone_strs.append( "  " + _gen_list_clone_strs( type_, f'self.{name}' ) + "," )

  return _create_fn(
    '__deepcopy__',
    [ 'self', 'memo' ],
    clone_strs + [ ')' ],
  )
#-------------------------------------------------------------------------
# _check_valid_array
#-------------------------------------------------------------------------

def _recursive_check_array_types( current ):
  x = current[0]
  if isinstance( x, list ):
    x_len  = len(x)
    x_type = _recursive_check_array_types( x )
    for y in current[1:]:
      assert isinstance( y, list ) and len(y) == x_len
      y_type = _recursive_check_array_types( y )
      assert y_type is x_type
    return x_type

  assert issubclass( x, Bits ) or is_bitstruct_class( x )
  for y in current[1:]:
    assert y is x
  return x

def _check_valid_array_of_types( arr ):
  # Check if the provided list is a strict multidimensional array
  try:
    return _recursive_check_array_types( arr )
  except Exception as e:
    print(e)
    return None

#-------------------------------------------------------------------------
# _check_field_annotation
#-------------------------------------------------------------------------

def _check_field_annotation( cls, name, type_ ):
  # Make sure not default is annotated
  if hasattr( cls, name ):
    default = getattr( cls, name )
    raise TypeError( "We don't allow subfields to have default value:\n"
                     f"- Field '{name}' of BitStruct {cls.__name__} has default value {default!r}." )

  # Special case if the type is an instance of list
  if isinstance( type_, list ):
    if _check_valid_array_of_types( type_ ) is None:
      raise TypeError( "The provided list spec should be a strict multidimensional ARRAY "
                       f"with no varying sizes or types. All non-list elements should be VALID types." )
  else:
    # Now we work with types
    if not isinstance( type_, type ):
      raise TypeError(f"{type_} is not a type\n"\
                      f"- Field '{name}' of BitStruct {cls.__name__} is annotated as {type_}.")

    # More specifically, Bits and BitStruct
    if not issubclass( type_, Bits ) and not is_bitstruct_class( type_ ):
      raise TypeError( "We currently only support BitsN, list, or another BitStruct as BitStruct field:\n"
                      f"- Field '{name}' of BitStruct {cls.__name__} is annotated as {type_}." )

#-------------------------------------------------------------------------
# _get_self_name
#-------------------------------------------------------------------------
# Return a self name based on fields.

def _get_self_name( fields ):
  return( _ANTI_CONFLICT_SELF_NAME if _DEFAULT_SELF_NAME in fields else
          _DEFAULT_SELF_NAME )

#-------------------------------------------------------------------------
# _process_cls
#-------------------------------------------------------------------------
# Process the input cls and add methods to it.

_bitstruct_hash_cache = {}

def _process_class( cls, add_init=True, add_str=True, add_repr=True,
                    add_hash=True ):

  # Get annotations of the class
  cls_annotations = cls.__dict__.get('__annotations__', {})
  if not cls_annotations:
    raise AttributeError( "No field is declared in the bit struct definition.\n"
                         f"Suggestion: check the definition of {cls.__name__} to"
                          " make sure it only contains 'field_name(string): Type(type).'" )

  # Get field information from the annotation and prepare for hashing
  fields = {}
  hashable_fields = {}

  def _convert_list_to_tuple( x ):
    if isinstance( x, list ):
      return tuple( [ _convert_list_to_tuple( y ) for y in x ] )
    return x

  for a_name, a_type in cls_annotations.items():
    _check_field_annotation( cls, a_name, a_type )
    fields[ a_name ] = a_type
    hashable_fields[ a_name ] = _convert_list_to_tuple( a_type )

  cls._hash = _hash = hash( (cls.__name__, *tuple(hashable_fields.items()),
                             add_init, add_str, add_repr, add_hash) )

  if _hash in _bitstruct_hash_cache:
    return _bitstruct_hash_cache[ _hash ]

  _bitstruct_hash_cache[ _hash ] = cls

  # Stamp the special attribute so that translation pass can identify it
  # as bit struct.
  setattr( cls, _FIELDS, fields )

  # Add methods to the class

  # Create __init__. Here I follow the dataclass convention that we only
  # add our generated __init__ function when add_init is true and user
  # did not define their own init.
  if add_init:
    if not '__init__' in cls.__dict__:
      cls.__init__ = _mk_init_fn( _get_self_name(fields), fields )

  # Create __str__
  if add_str:
    if not '__str__' in cls.__dict__:
      cls.__str__ = _mk_str_fn( fields )

  # Create __repr__
  if add_repr:
    if not '__repr__' in cls.__dict__:
      cls.__repr__ = _mk_repr_fn( fields )

  # Create __eq__. There is no need for a __ne__ method as python will
  # call __eq__ and negate it.
  # NOTE: if user overwrites __eq__ it may lead to different behavior for
  # the translated verilog as in the verilog world two bit structs are
  # equal only if all the fields are equal. We always try to add __eq__

  if not '__eq__' in cls.__dict__:
    cls.__eq__ = _mk_eq_fn( fields )
  else:
    w_msg = ( f'Overwriting {cls.__qualname__}\'s __eq__ may cause the '
              'translated verilog behaves differently from PyMTL '
              'simulation.')
    warnings.warn( w_msg )

  # Create __hash__.
  if add_hash:
    if not '__hash__' in cls.__dict__:
      cls.__hash__ = _mk_hash_fn( fields )

  # Shunning: add __ilshift__ and _flip for update_ff
  assert not '__ilshift__' in cls.__dict__ and not '_flip' in cls.__dict__

  cls.__ilshift__, cls._flip = _mk_ff_fn( fields )

  # Shunning: add clone
  assert not 'clone' in cls.__dict__ and not '__deepcopy__' in cls.__dict__

  cls.clone = _mk_clone_fn( fields )

  cls.__deepcopy__ = _mk_deepcopy_fn( fields )

  assert not 'get_field_type' in cls.__dict__

  def get_field_type( cls, name ):
    if name in cls.__bitstruct_fields__:
      return cls.__bitstruct_fields__[ name ]
    raise AttributeError( f"{cls} has no field '{name}'" )

  cls.get_field_type = classmethod(get_field_type)

  # TODO: maybe add a to_bits and from bits function.

  return cls

#-------------------------------------------------------------------------
# bitstruct
#-------------------------------------------------------------------------
# The actual class decorator. We add a * in the argument list so that the
# following argument can only be used as keyword arguments.

def bitstruct( _cls=None, *, add_init=True, add_str=True, add_repr=True, add_hash=True ):

  def wrap( cls ):
    return _process_class( cls, add_init, add_str, add_repr )

  # Called as @bitstruct(...)
  if _cls is None:
    return wrap

  # Called as @bitstruct without parens.
  return wrap( _cls )

#-------------------------------------------------------------------------
# mk_bitstruct
#-------------------------------------------------------------------------
# Dynamically generate a bit struct class.
# TODO: should we add base parameters to support inheritence?

def mk_bitstruct( cls_name, fields, *, namespace=None, add_init=True,
                   add_str=True, add_repr=True, add_hash=True ):

  # copy namespace since  will mutate it
  namespace = {} if namespace is None else namespace.copy()

  # We assume fields is a dictionary and thus there won't be duplicate
  # field names. So we only check if the field names are indeed strings
  # and that they are not keywords.
  annos = {}
  for name, f in fields.items():
    if not isinstance( name, str ) or not name.isidentifier():
      raise TypeError( f'Field name {name!r} is not a valid identifier!' )
    if keyword.iskeyword( name ):
      raise TypeError( f'Field name {name!r} is a keyword!' )
    annos[ name ] = f

  namespace['__annotations__'] = annos
  cls = types.new_class( cls_name, (), {}, lambda ns: ns.update( namespace ) )
  return bitstruct( cls, add_init=add_init, add_str=add_str,
                    add_repr=add_repr, add_hash=add_hash )



"""
#=========================================================================
# OpenLoopCLPass_test.py
#=========================================================================
#
# Author : Shunning Jiang
# Date   : Apr 19, 2019
"""
from pymtl3.datatypes import Bits32
from pymtl3.dsl import *
from pymtl3.dsl.errors import UpblkCyclicError
from pymtl3.passes import TracingConfigs
from pymtl3.passes.sim.GenDAGPass import GenDAGPass

from ..OpenLoopCLPass import OpenLoopCLPass


def test_top_level_method_tracing():

  class Top(Component):

    def construct( s ):
      s.element = None

      s.count = Wire(Bits32)
      s.amp   = Wire(Bits32)

      s.value = Wire(Bits32)

      @s.update_ff
      def up_incr():
        s.count <<= s.count + 1

      @s.update
      def up_amp():
        s.amp = s.count * 100

      @s.update
      def up_compose_in():
        if s.element:
          s.value = s.amp + s.element
          s.element = None
        else:
          s.value = Bits32( -1 )

      s.add_constraints(
        M( s.push ) < U( up_compose_in ),
        U( up_compose_in ) < M( s.pull ), # bypass behavior
      )

    @method_port
    def push( s, ele ):
      if s.element is None:
        s.element = ele

    @method_port
    def pull( s ):
      return s.value

    def line_trace( s ):
      return "line trace: {}".format(s.value)

    def done( s ):
      return True

  A = Top()
  A.elaborate()

  # Turn on textwave
  A.config_tracing = TracingConfigs( tracing='text_fancy' )

  A.apply( GenDAGPass() )
  A.apply( OpenLoopCLPass() )
  A.lock_in_simulation()

  print("- push!")
  A.push(7)
  print("- pull!")
  print(A.pull())

  print("- pull!")
  print(A.pull())

  print("- push!")
  A.push(33)
  print("- push!")
  A.push(55)
  print("- pull!")
  print(A.pull())

  print("num_cycles_executed: ", A.num_cycles_executed)
  A.print_textwave()

class TestModuleNonBlockingIfc(Component):

  def construct( s ):
    s.element = None

    s.count = Wire(Bits32)
    s.amp   = Wire(Bits32)

    s.value = Wire(Bits32)

    @s.update_ff
    def up_incr():
      s.count <<= s.count + 1

    @s.update
    def up_amp():
      s.amp = s.count * 100

    @s.update
    def up_compose_in():
      if s.element:
        s.value = s.amp + s.element
        s.element = None
      else:
        s.value = -1

    s.add_constraints(
      M( s.push ) < U( up_compose_in ),
      U( up_compose_in ) < M( s.pull ), # bypass behavior
    )

  @non_blocking( lambda s: s.element is None and s.count % 5 == 4 )
  def push( s, ele ):
    assert s.element is None and s.count % 5 == 4
    s.element = ele

  @non_blocking( lambda s: s.value >= 0 )
  def pull( s ):
    return s.value

  def line_trace( s ):
    return "line trace: {}".format(s.value)

  def done( s ):
    return True

def _test_TestModuleNonBlockingIfc( cls ):

  A = cls()
  A.elaborate()
  A.apply( GenDAGPass() )
  A.apply( OpenLoopCLPass() )
  A.lock_in_simulation()

  rdy = A.push.rdy()
  print("- push_rdy?", rdy )
  assert not rdy

  rdy = A.push.rdy()
  print("- push_rdy?", rdy)
  assert not rdy

  rdy = A.push.rdy()
  print("- push_rdy?", rdy)
  assert not rdy

  rdy = A.push.rdy()
  print("- push_rdy?", rdy)
  assert not rdy

  rdy = A.push.rdy()
  print("- push_rdy?", rdy)
  assert rdy
  print("- push 13!")
  A.push(13)

  assert A.pull.rdy()
  ret = A.pull()
  print("- pull!", ret)
  assert ret == 413

  assert not A.pull.rdy()

  rdy = A.push.rdy()
  print("- push_rdy?", rdy )
  assert not rdy

  rdy = A.push.rdy()
  print("- push_rdy?", rdy)
  assert not rdy

  rdy = A.push.rdy()
  print("- push_rdy?", rdy)
  assert not rdy

  rdy = A.push.rdy()
  print("- push_rdy?", rdy)
  assert rdy
  print("- push 33!")
  A.push(33)

  assert A.pull.rdy()
  ret = A.pull()
  print("- pull!", ret)
  assert ret == 933

  assert not A.pull.rdy()

  return A.num_cycles_executed

def test_top_level_non_blocking_ifc():
  num_cycles = _test_TestModuleNonBlockingIfc( TestModuleNonBlockingIfc )
  assert num_cycles == 10 # regression

def test_top_level_non_blocking_ifc_in_deep_net():

  class Top(Component):
    def construct( s ):
      s.push = CalleeIfcCL()
      s.pull = CalleeIfcCL()
      s.inner = Top_less_less_inner()( push = s.push, pull = s.pull )
    def line_trace( s ):
      return s.inner.line_trace()

    def done( s ):
      return True

  class Top_less_less_inner(Component):
    def construct( s ):
      s.push = CalleeIfcCL()
      s.pull = CalleeIfcCL()
      s.inner = Top_less_inner()( push = s.push, pull = s.pull )
    def line_trace( s ):
      return s.inner.line_trace()

    def done( s ):
      return True

  class Top_less_inner(Component):
    def construct( s ):
      s.push = CalleeIfcCL()
      s.pull = CalleeIfcCL()
      s.inner = TestModuleNonBlockingIfc()( push = s.push, pull = s.pull )

    def line_trace( s ):
      return s.inner.line_trace()

    def done( s ):
      return True

  num_cycles = _test_TestModuleNonBlockingIfc( Top )
  assert num_cycles == 10 # regression

class PassThroughPlus100( Component ):

  @non_blocking( lambda s: s.real_push.rdy() )
  def push( s, msg ):
    assert s.real_push.rdy()
    s.real_push( msg )

  def construct( s ):
    s.real_push = CallerIfcCL()

    s.add_constraints(
      M(s.push) == M(s.real_push),
    )

def test_pass_through_equal_m_constraint():

  class Top(Component):
    def construct( s ):
      s.push = CalleeIfcCL()
      s.pull = CalleeIfcCL()
      s.pass1 = PassThroughPlus100()( push = s.push )
      s.inner = TestModuleNonBlockingIfc()( push = s.pass1.real_push, pull = s.pull )

    def line_trace( s ):
      return s.inner.line_trace()

    def done( s ):
      return True

  num_cycles = _test_TestModuleNonBlockingIfc( Top )
  assert num_cycles == 10 # regression


def test_deep_pass_through_equal_m_constraint():

  class Top(Component):
    def construct( s ):
      s.push = CalleeIfcCL()
      s.pull = CalleeIfcCL()
      s.through = [ PassThroughPlus100() for _ in range(10) ]
      for i in range(10):
        connect( s.through[i].push, s.push if i == 0 else \
                                      s.through[i-1].real_push,
        )

      s.inner = TestModuleNonBlockingIfc()(
        push = s.through[-1].real_push,
        pull = s.pull,
      )

    def line_trace( s ):
      return "push's line trace:" + str(s.push)

    def done( s ):
      return True

  num_cycles = _test_TestModuleNonBlockingIfc( Top )
  assert num_cycles == 10 # regression



#=========================================================================
# StructuralTranslatorL4.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 4, 2019
"""Provide L4 structural translator."""

from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir import StructuralRTLIRSignalExpr as sexp
from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL4Pass import (
    StructuralRTLIRGenL4Pass,
)

from .StructuralTranslatorL3 import StructuralTranslatorL3


class StructuralTranslatorL4( StructuralTranslatorL3 ):
  #-----------------------------------------------------------------------
  # gen_structural_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def gen_structural_trans_metadata( s, tr_top ):
    tr_top.apply( StructuralRTLIRGenL4Pass( s.inst_conns ) )

  #-----------------------------------------------------------------------
  # translate_structural
  #-----------------------------------------------------------------------

  # Override
  def translate_structural( s, tr_top ):
    s.structural.decl_subcomps = {}
    super().translate_structural( tr_top )

  #-----------------------------------------------------------------------
  # _translate_structural
  #-----------------------------------------------------------------------

  # Override
  def _translate_structural( s, m ):
    super()._translate_structural( m )
    for child in m.get_child_components():
      s._translate_structural( child )

  # Override
  def translate_decls( s, m ):
    super().translate_decls( m )
    m_rtype = m._pass_structural_rtlir_gen.rtlir_type

    # Translate subcomponent declarations
    subcomp_decls = []
    for c_id, _c_rtype in m_rtype.get_subcomps_packed():
      if isinstance( _c_rtype, rt.Array ):
        c_array_rtype = _c_rtype
        c_rtype = _c_rtype.get_sub_type()
      else:
        c_array_rtype = None
        c_rtype = _c_rtype

      # Translate ports of the subcomponent
      port_conns, ifc_conns = [], []
      for port_id, _port_rtype in c_rtype.get_ports_packed():
        if isinstance( _port_rtype, rt.Array ):
          port_array_rtype = _port_rtype
          port_rtype = _port_rtype.get_sub_type()
        else:
          port_array_rtype = None
          port_rtype = _port_rtype

        port_conns.append( s.rtlir_tr_subcomp_port_decl(
          m,
          c_id, c_rtype, s.rtlir_tr_unpacked_array_type( c_array_rtype ),
          port_id, port_rtype,
          s.rtlir_data_type_translation( m, port_rtype.get_dtype() ),
          s.rtlir_tr_unpacked_array_type( port_array_rtype )
        ) )

      # Translate interfaces of the subcomponent
      for ifc_port_id, _ifc_port_rtype in c_rtype.get_ifc_views_packed():
        if isinstance( _ifc_port_rtype, rt.Array ):
          ifc_port_array_rtype = _ifc_port_rtype
          ifc_port_rtype = _ifc_port_rtype.get_sub_type()
        else:
          ifc_port_array_rtype = None
          ifc_port_rtype = _ifc_port_rtype

        ports = []
        all_ifc_ports = ifc_port_rtype.get_all_properties_packed()
        for port_id, p_rtype in all_ifc_ports:
          if isinstance( p_rtype, rt.Array ):
            port_array_rtype = p_rtype
            port_rtype = p_rtype.get_sub_type()
          else:
            port_array_rtype = None
            port_rtype = p_rtype

          # Translate a single port of the current interface
          ports.append( s.rtlir_tr_subcomp_ifc_port_decl(
            m,
            '{c_id}', c_rtype, s.rtlir_tr_unpacked_array_type( c_array_rtype ),
            ifc_port_id, ifc_port_rtype,
            s.rtlir_tr_unpacked_array_type( ifc_port_array_rtype ),
            port_id, port_rtype,
            s.rtlir_tr_unpacked_array_type( port_array_rtype )
          ) )

        # Assemble all ports of the current interface into a complete interface
        ifc_conns.append( s.rtlir_tr_subcomp_ifc_decl(
          m,
          '{c_id}', c_rtype, s.rtlir_tr_unpacked_array_type( c_array_rtype ),
          ifc_port_id, ifc_port_rtype,
          s.rtlir_tr_unpacked_array_type( ifc_port_array_rtype ),
          s.rtlir_tr_subcomp_ifc_port_decls( ports )
        ) )

      # Generate a list of ports and interfaces
      subcomp_decls.append( s.rtlir_tr_subcomp_decl(
        m,
        c_id, c_rtype, s.rtlir_tr_unpacked_array_type( c_array_rtype ),
        s.rtlir_tr_subcomp_port_decls( port_conns ),
        s.rtlir_tr_subcomp_ifc_decls( ifc_conns )
      ) )
    s.structural.decl_subcomps[m] = s.rtlir_tr_subcomp_decls( subcomp_decls )

  #-----------------------------------------------------------------------
  # rtlir_signal_expr_translation
  #-----------------------------------------------------------------------

  def rtlir_signal_expr_translation( s, expr, m ):
    """Translate a signal expression in RTLIR into its backend representation.

    Add support for the following operations at L4: sexp.SubCompAttr
    """
    if isinstance( expr, sexp.SubCompAttr ):
      return s.rtlir_tr_subcomp_attr(
        s.rtlir_signal_expr_translation( expr.get_base(), m ),
        expr.get_attr() )

    elif isinstance( expr, sexp.ComponentIndex ):
      return s.rtlir_tr_component_array_index(
        s.rtlir_signal_expr_translation( expr.get_base(), m ),
        expr.get_index() )

    else:
      return super(). \
              rtlir_signal_expr_translation( expr, m )

  #-----------------------------------------------------------------------
  # Methods to be implemented by the backend
  #-----------------------------------------------------------------------

  # Declarations
  def rtlir_tr_subcomp_port_decls( s, port_decls ):
    raise NotImplementedError()

  def rtlir_tr_subcomp_port_decl( s, m, c_id, c_rtype, c_array_type, port_id, port_rtype,
      port_dtype, port_array_type ):
    raise NotImplementedError()

  def rtlir_tr_subcomp_ifc_decls( s, ifc_decls ):
    raise NotImplementedError()

  def rtlir_tr_subcomp_ifc_decl( s, m, c_id, c_rtype, c_array_type, ifc_port_id,
      ifc_port_rtype, ifc_port_array_type, ports ):
    raise NotImplementedError()

  def rtlir_tr_subcomp_ifc_port_decls( s, ifc_port_decls ):
    raise NotImplementedError()

  def rtlir_tr_subcomp_ifc_port_decl( s, m, c_id, c_rtype, c_array_type,
      ifc_id, ifc_rtype, ifc_array_rtype, port_id, port_rtype,
      port_array_type ):
    raise NotImplementedError()

  def rtlir_tr_subcomp_decls( s, subcomps ):
    raise NotImplementedError()

  def rtlir_tr_subcomp_decl( s, m, c_id, c_rtype, c_array_type, port_conns, ifc_conns ):
    raise NotImplementedError()

  # Signal operations
  def rtlir_tr_component_array_index( s, base_signal, index ):
    raise NotImplementedError()

  def rtlir_tr_subcomp_attr( s, base_signal, attr ):
    raise NotImplementedError()



"""
========================================================================
UnrollTickPass.py
========================================================================
Generate an unrolled tick function

Author : Shunning Jiang
Date   : Dec 26, 2018
"""

import py

from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.errors import PassOrderError
from pymtl3.passes.sim.SimpleTickPass import SimpleTickPass


class UnrollTickPass( BasePass ):

  @staticmethod
  def gen_tick_function( schedule ):

    # Berkin IlBeyi's recipe ( updated using f-strings and enumerate )
    strs = [f"  update_blk{idx}() # {sched}" for idx, sched in \
        enumerate([ x.__name__ for x in schedule ])]

    gen_tick_src = """
        def compile_unroll( schedule ):
          {}
          def tick_unroll():
            # The code below does the actual calling of update blocks.
          {}
          return tick_unroll
        """.format( ";".join( map(
                        "update_blk{0}=schedule[{0}]".format,
                        range( len( schedule ) ) ) ),
                        "\n          ".join( strs ) )

    l = {}
    exec(py.code.Source( gen_tick_src ).compile(), l)
    return l['compile_unroll']( schedule )

  def __call__( self, top ):
    if not hasattr( top, "_sched" ):
      raise PassOrderError( "_sched" )
    if not hasattr( top._sched, "update_schedule" ):
      raise PassOrderError( "update_schedule" )
    if not hasattr( top._sched, "schedule_ff" ):
      raise PassOrderError( "schedule_ff" )
    if not hasattr( top._sched, "schedule_posedge_flip" ):
      raise PassOrderError( "schedule_posedge_flip" )

    # We assemble the final schedule from multiple sources of required
    # work to generate a tick function for simulation

    # Currently the tick order is:
    # [ ff, tracing, posedge, clear_cl_trace, update ]

    final_schedule = []
    # call ff blocks first
    final_schedule.extend( top._sched.schedule_ff )

    # no tracing

    # posedge flip
    final_schedule.extend( top._sched.schedule_posedge_flip )

    # advance cycle after posedge
    def generate_advance_sim_cycle( top ):
      def advance_sim_cycle():
        top.simulated_cycles += 1
      return advance_sim_cycle
    final_schedule.append( generate_advance_sim_cycle(top) )

    # execute all update blocks
    final_schedule.extend( top._sched.update_schedule )

    top.tick = UnrollTickPass.gen_tick_function( final_schedule )
    # reset sim_cycles
    top.simulated_cycles = 0
